"""
This class generates an EAN bar code image from a barcode string.
If the barcode string has no checksum, it will be calculated and
added; if it already has a checksum, it will be verified.

It requires PIL (python imaging library) installed.

Create bar code sample:
   from EANBarCode import EanBarCode
   bar = EanBarCode()
   bar.getImage("9782212110708",50,"png")

Original code by remi inconnu
http://code.activestate.com/recipes/426069/

Improvements by A. Digulla:
- Got rid of ImageFont and external font files. Instead, the shapes of the
  digits are copied from an image that is created in RAM.
- Fixed typos
- Replaced string concatenation in makeCode with StringIO buffer
- getImage() no longer writes an image file to disk; it only returns it.
- Added test data for the image tests to verify that images are generated
  correctly.
"""

from cStringIO import StringIO
import Image

# An image which contains a strip of digits. Each digit is 6x9 pixels.
__numbers_base64 = '''
iVBORw0KGgoAAAANSUhEUgAAAGQAAAAKCAAAAABnd6SiAAAA70lEQVR4nGL8z0B7AAAAAP//YqKD
HQwAAAAA//+MyjEKACAQxMBZ8P9fXgsPsTRVCFlCETU+RTTQU16/zxcbAAD//2JiZPjPwMjAyAg1
lYERSjJADGX4//8/XOQ/hhriAAAAAP//YoFQCEPgLmT8D7cYIkh+7AEAAAD//0KLE0igIbsT6iu4
hf/hJPEAAAAA//9CtQTi7P/w4GL8/x/ZKpgSRgaG/6T4CwAAAP//QrGEkeE/Iyx+/v//zwCNeFQn
QJikeIQBAAAA//9i/M/IgOr5/wxoqQsp1UFinuTUBQAAAP//YqRHZgQAAAD//wMA2dk9Gcsv5nUA
AAAASUVORK5CYII=
'''

def imageFromBase64String(s):
    from base64 import decodestring
    
    fh = StringIO(decodestring(s.strip()))
    img = Image.open(fh)
    img.load()
    fh.close()
    return img


__numbers_img = None
def numbers_img():
    # Create the image strip of digits
    global __numbers_img
    
    if __numbers_img is None:
        __numbers_img = imageFromBase64String(__numbers_base64)
    
    return __numbers_img

class EanBarCode:
    """ Compute the EAN bar code """
    def __init__(self):
        A = {0 : "0001101", 1 : "0011001", 2 : "0010011", 3 : "0111101", 4 : "0100011", 
             5 : "0110001", 6 : "0101111", 7 : "0111011", 8 : "0110111", 9 : "0001011"}
        B = {0 : "0100111", 1 : "0110011", 2 : "0011011", 3 : "0100001", 4 : "0011101",
             5 : "0111001", 6 : "0000101", 7 : "0010001", 8 : "0001001", 9 : "0010111"}
        C = {0 : "1110010", 1 : "1100110", 2 : "1101100", 3 : "1000010", 4 : "1011100",
             5 : "1001110", 6 : "1010000", 7 : "1000100", 8 : "1001000", 9 : "1110100"}
        self.groupC = C
        
        self.family = {0 : (A,A,A,A,A,A), 1 : (A,A,B,A,B,B), 2 : (A,A,B,B,A,B), 3 : (A,A,B,B,B,A), 4 : (A,B,A,A,B,B),
                       5 : (A,B,B,A,A,B), 6 : (A,B,B,B,A,A), 7 : (A,B,A,B,A,B), 8 : (A,B,A,B,B,A), 9 : (A,B,B,A,B,A)}


    def makeCode(self, code):
        """ Create the binary code
        return a string which contains "0" for white bar, "1" for black bar, "L" for long bar
        """
        
        # Convert code string in integer list
        self.EAN13 = []
        for digit in code:
            self.EAN13.append(int(digit))
        
        # If the code has already a checksum
        if len(self.EAN13) == 13:
            # Verify checksum
            self.verifyChecksum(self.EAN13)
        # If the code has no checksum, yet
        elif len(self.EAN13) == 12:
            # Add checksum value
            self.EAN13.append(self.computeChecksum(self.EAN13))
        
        # Get the left codage class
        left = self.family[self.EAN13[0]]
        
        # Add start separator
        strCode = StringIO()
        strCode.write('L0L')
        
        # Compute the left part of bar code
        for i in range(0,6):
            strCode.write(left[i][self.EAN13[i+1]])
        
        # Add middle separator 
        strCode.write('0L0L0')
        
        # Compute the right codage class
        for i in range (7,13):
            strCode.write(self.groupC[self.EAN13[i]])
        
        # Add stop separator
        strCode.write('L0L')
        
        return strCode.getvalue()


    def computeChecksum(self, arg):
        """ Compute the checksum of bar code """
        # UPCA/EAN13
        weight=[1,3]*6
        magic=10
        sum = 0
        
        for i in range(12):         # checksum based on first 12 digits.
            sum += int(arg[i]) * weight[i]
        z = ( magic - (sum % magic) ) % magic
        if z < 0 or z >= magic:
            return None
        return z


    def verifyChecksum(self, bits): 
        """ Verify the checksum """
        computedChecksum = self.computeChecksum(bits[:12])
        codeBarChecksum = bits[12]
        
        if codeBarChecksum != computedChecksum:
            raise Exception ("Bad checksum is %s and should be %s"%(codeBarChecksum, computedChecksum))


    def getImage(self, value, height = 50):
        """ Get an image with PIL library 
        value code barre value
        height height in pixel of the bar code
        extension image file extension"""
        #import Image, ImageFont, ImageDraw
        import Image, ImageDraw
        
        # Get the bar code list
        bits = self.makeCode(value)
        
        # Create barcode string with checksum
        code = ""
        for digit in self.EAN13:
           code += "%d"%digit
        
        # Create a new image
        position = 8
        im = Image.new("1",(len(bits)+position,height),255)
        
        # Load font
        #font = ImageFont.load("courB08.pil")
        
        #im2 = Image.new("1", (10*10, 10),255)
        #draw2 = ImageDraw.Draw(im2)
        #draw2.rectangle(((0,0),(im.size[0],im.size[1])),fill=256)
        #draw2.text((0, 0), '0123456789', font=font, fill=0)
        im2 = numbers_img()
        #im2.save("numbers.png", "PNG")
        
        #fh = open('numbers.png', 'rb')
        #data = fh.read()
        #fh.close()
        #print len(data)
        #from base64 import encodestring
        #print encodestring(data)
        
        def stamp(dest, pos, digit):
             digit = int(digit)
             offset = digit*6
             #print offset
             
             copy = im2.crop((offset, 0, offset+6, 9))
             dest.paste(copy, pos)
        
        # Create drawer
        draw = ImageDraw.Draw(im)
        
        # Erase image
        #draw.rectangle(((0,0),(im.size[0],im.size[1])),fill=256)
        
        # Draw first part of number
        #draw.text((0, height-9), code[0], font=font, fill=0)
        stamp(im, (0, height-9), code[0])
        
        # Draw first part of number
        #draw.text((position+7, height-9), code[1:7], font=font, fill=0)
        for i in range(1,7):
            stamp(im, (position+7+(i-1)*6, height-9), code[i])
        
        # Draw second part of number
        #draw.text((len(bits)/2+6+position, height-9), code[7:], font=font, fill=0)
        for i in range(0,6):
            stamp(im, (len(bits)/2+6+position+i*6, height-9), code[i+7])
        
        # Draw the bar codes
        position -= 1
        for bit in bits:
            position += 1
            if bit == '0':
                continue
            
            # Draw long bar if bit is 'L'
            y = height-3 if bit == 'L' else height-10
            draw.rectangle(((position,0),(position,y)),fill=0)
        
        # Save the result image
        #im.save(code+"."+extension.lower(), extension.upper())
        return (code, im)


#def decodeFontFile(data, file):
#   """ Decode font file embedded in this script and create file """
#   from zlib import decompress
#   from base64 import decodestring
#   from os.path import exists
#   
#   # If the font file is missing
#   if not exists(file):
#      # Write font file
#      open (file, "wb").write(decompress(decodestring(data)))


def testWithChecksum():
    """ Test bar code with checksum """
    bar = EanBarCode()
    assert(bar.makeCode('0000000000000')== 'L0L0001101000110100011010001101000110100011010L0L0111001011100101110010111001011100101110010L0L' )
    assert(bar.makeCode('1111111111116')== 'L0L0011001001100101100110011001011001101100110L0L0110011011001101100110110011011001101010000L0L' )
    assert(bar.makeCode('2222222222222')== 'L0L0010011001001100110110011011001001100110110L0L0110110011011001101100110110011011001101100L0L' )
    assert(bar.makeCode('3333333333338')== 'L0L0111101011110101000010100001010000101111010L0L0100001010000101000010100001010000101001000L0L' )
    assert(bar.makeCode('4444444444444')== 'L0L0100011001110101000110100011001110100111010L0L0101110010111001011100101110010111001011100L0L' )
    assert(bar.makeCode('5555555555550')== 'L0L0110001011100101110010110001011000101110010L0L0100111010011101001110100111010011101110010L0L' )
    assert(bar.makeCode('6666666666666')== 'L0L0101111000010100001010000101010111101011110L0L0101000010100001010000101000010100001010000L0L' )
    assert(bar.makeCode('7777777777772')== 'L0L0111011001000101110110010001011101100100010L0L0100010010001001000100100010010001001101100L0L' )
    assert(bar.makeCode('8888888888888')== 'L0L0110111000100101101110001001000100101101110L0L0100100010010001001000100100010010001001000L0L' )
    assert(bar.makeCode('9999999999994')== 'L0L0001011001011100101110001011001011100010110L0L0111010011101001110100111010011101001011100L0L' )   


def testWithoutChecksum():
    """ Test bar code without checksum """
    bar = EanBarCode()
    assert(bar.makeCode('000000000000')== 'L0L0001101000110100011010001101000110100011010L0L0111001011100101110010111001011100101110010L0L' )
    assert(bar.makeCode('111111111111')== 'L0L0011001001100101100110011001011001101100110L0L0110011011001101100110110011011001101010000L0L' )
    assert(bar.makeCode('222222222222')== 'L0L0010011001001100110110011011001001100110110L0L0110110011011001101100110110011011001101100L0L' )
    assert(bar.makeCode('333333333333')== 'L0L0111101011110101000010100001010000101111010L0L0100001010000101000010100001010000101001000L0L' )
    assert(bar.makeCode('444444444444')== 'L0L0100011001110101000110100011001110100111010L0L0101110010111001011100101110010111001011100L0L' )
    assert(bar.makeCode('555555555555')== 'L0L0110001011100101110010110001011000101110010L0L0100111010011101001110100111010011101110010L0L' )
    assert(bar.makeCode('666666666666')== 'L0L0101111000010100001010000101010111101011110L0L0101000010100001010000101000010100001010000L0L' )
    assert(bar.makeCode('777777777777')== 'L0L0111011001000101110110010001011101100100010L0L0100010010001001000100100010010001001101100L0L' )
    assert(bar.makeCode('888888888888')== 'L0L0110111000100101101110001001000100101101110L0L0100100010010001001000100100010010001001000L0L' )
    assert(bar.makeCode('999999999999')== 'L0L0001011001011100101110001011001011100010110L0L0111010011101001110100111010011101001011100L0L' )   


def assertEquals(code, expected, im):
    """ Compare two images.
    
    code: If the compare fails, both images will be written to disk
        as PNG images. This is the base name of the two files.
    expected: The expected result either as an image or as a BASE64
        encoded string.
    im: The image to check
    """
    
    if type(expected) == type(''):
        expected = imageFromBase64String(expected)
    s1 = expected.getdata()
    s2 = im.getdata()
    width = expected.size[0]
    for i in range(len(s1)):
        if s1[i] != s2[i]:
            expected.save(code+'-expected.png')
            im.save(code+'.png')
            from base64 import encodestring
            s = encodestring(fh.getvalue())
            print s
            
            x = i%width
            y = i/width
            assert False, 'Images differ at %d/%d' % (x,y)


def testImage():
    """ Test images generation with PIL """
    bar = EanBarCode()
    #code, im = bar.getImage("9782212110708",50)
    #im.save(code+'.gif', 'GIF')
    # Note the size difference
    #im.save(code+'.png', 'PNG')
    
    expected = '''iVBORw0KGgoAAAANSUhEUgAAAGcAAAAUAQAAAAC4jB47AAABHElEQVR4nGL8H3PM0dqv65lBv5f5
TQAAAAD//2JiQAYAAAAA//9C5QEAAAD//0LlAQAAAP//QuUBAAAA//9C5QEAAAD//0LlAQAAAP//
QuUBAAAA//9C5QEAAAD//0LlAQAAAP//QuUBAAAA//9i+B///////3/r//////8LAAAA//9ClQMA
AAD//2K5wCCbwD5V1b4o7OrVZgAAAAD//2Jy+K+7R9qQ4e/m07WbGQAAAAD//2I6wNj254UZAwMD
YxsDAwAAAP//YmJgOGvAM9+B2Yfhvy8DAAAA//9iYWBY/JDPgOHvGeV/xQwAAAAA//9iYpjPJePw
r6nhBcNfBy4AAAAA//9iMmE47HD1AgOzr3E6LwMAAAD//wMABcczVq45mHwAAAAASUVORK5CYII='''
    
    code, im = bar.getImage("123456789012",20)
    assertEquals(code, expected, im)
    
    #code, im = bar.getImage("200100010001",50)
    #im.save(code+'.png', 'PNG')


def test():
    """ Execute all tests """
    testWithChecksum()
    testWithoutChecksum()
    testImage()


if __name__ == "__main__":
    import sys
    if len(sys.argv) == 1:
        test()
    else:
        bar = EanBarCode()
        bar.getImage(sys.argv[1],50,"png")
        
