#!/usr/bin/env python
# -*- coding=iso-8859-1 -*-
"""
A little tool to download product information from Amazon.

The tool has three modes:

1. No parameters
2. A list of product codes as parameters
3. '-t'

In mode 1, the tool will read product codes from the standard input. Press
RETURN (without a code) or Ctrl-D to exit. If PyQt4 is available, it
will open a nice UI to scan and manage your property.

In mode 2, it will use the product codes supplied on the command line.
After fetching all those, it will exit.

In mode 3, it will run a couple of tests.

If a product can't be identified, its barcode will be appended to the file
"failed-codes".
"""

# TODOs:
#
# - Show original response in debug dialog
# - Allow to search Google when Amazon can't find anything
# - Show details of article (searchIndex, etc)
# - Remember the last location
# - Remember the UI settings (column width, window site+position, etc)
# - Allow to change and remember the order of the table columns
# - Load article table in chunks
# - Update id2Location when new locations are added
# - Import data from XML
# - Dialog to manage locations
# - Move location table into a new dialog
# - Button to generate an OpenOffice document with barcodes
# - Button to export to OpenOffice spreadsheet
# - Rows in table are too heigh
# - Documentation
# - Allow to manually call Amazon and merge the result
# - If nothing can be found, add a dummy entry
# - Allow to add arbitrary additional data to existing entries
# - Allow to create new locations
# - Fetch data in background thread
# - Allow to edit series/issue info
# - Import/Export/merge series/issue info
#
# Done:
#
# - Update the DB in a background thread while the user types
# - Export data as XML
# - CoverDialog can now (down-)load images and assign them to articles
# - i18n
# - Load cover image in cancelable background thread
# - Load articles in a background thread
# 0.7:
# - Filter non-digits from scanner input
# - If an entry already exists, select it after the scan for edit
# - Add series/issue info
# - Add a column for "Series" and "Issue" information (for collectors)
# - Allow to manually add an item to the list
# - Add lending information
# - Delete article(s)
# - Implement switchLocation

from __future__ import with_statement
from httplib import HTTPConnection
import httplib
from urllib import urlencode
import urllib2
from xml.etree.cElementTree import XMLTreeBuilder, dump, tostring, ElementTree
import sqlite3
import os
import codecs
import sys
import logging
import ConfigParser
from barcode import EanBarCode
import types
from threading import Thread, Event
from Queue import Queue
import base64
from StringIO import StringIO
import re

VERSION = '0.7'

config = ConfigParser.ConfigParser()
config.read(['upcscan.conf'])

if not config.has_section('upcscan'):
    config.add_section('upcscan')

AWSAccessKeyId = None
dbName = None
if config.has_option('upcscan', 'AWSAccessKeyId'):
    AWSAccessKeyId = config.get('upcscan', 'AWSAccessKeyId')
if config.has_option('upcscan', 'dbName'):
    dbName = config.get('upcscan', 'dbName')

if not dbName:
    dbName = 'articles.dat'
    config.set('upcscan', 'dbName', dbName)

if not AWSAccessKeyId or AWSAccessKeyId == 'XXXXXXXXXXXXXXXXXXXX':
    config.set('upcscan', 'AWSAccessKeyId', 'XXXXXXXXXXXXXXXXXXXX')
    with open('upcscan.conf', 'w') as fh:
        config.write(fh)
        raise Exception("You must specify a valid AWSAccessKeyId in upcscan.conf. If you don't have one, yet, request it at http://amazon.com/soap")

ISO_8859_1 = 'iso-8859-1'

# set up logging to file - see previous section for more details
logging.basicConfig(level=logging.DEBUG,
                    format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
                    datefmt='%m-%d %H:%M',
                    filename='scan.log',
                    filemode='a')
# define a Handler which writes INFO messages or higher to the sys.stderr
console = logging.StreamHandler()
console.setLevel(logging.DEBUG)
# set a format which is simpler for console use
formatter = logging.Formatter('%(name)-12s: %(levelname)-8s %(message)s')
# tell the handler to use this format
console.setFormatter(formatter)
# add the handler to the root logger
logging.getLogger('').addHandler(console)

# FUCKING PYTHON UNICODE MESS!
import locale
sys.stdout = codecs.getwriter(locale.getpreferredencoding())(sys.stdout)
#line = u"\u0411\n"
#sys.stdout.write(line)
#print line

def tr(*args):
    s = unicode(args[0])
    if u'%' not in s:
        return s
    
    buf = StringIO()
    pos = 0
    N = len(s)
    #print 'tr',repr(s)
    while pos < N:
        pos2 = s.find('%', pos)
        #print pos2, s[pos2:]
        if pos2 == -1:
            pos2 = N
        
        buf.write(s[pos:pos2])
        if pos2 == N:
            break
        if pos2+1 == N:
            buf.write('%')
            break
        
        c = s[pos2+1]
        if c not in '0123456789':
            buf.write('%')
            pos = pos2+1
            continue
        
        if pos2+2 < N and s[pos2+2] in '0123456789':
            c = s[pos2+1:pos2+3]
        
        i = int(c)
        buf.write(args[i])
        pos = pos2 + len(c) + 1
    return buf.getvalue()

def tr2(context, *args):
    return tr(*args)

class Service(object):
    NO_DIGITS = re.compile(r'\D', re.UNICODE)
    
    def __init__(self, server, port=80, url='/index.html', name=u''):
        self.server = server
        self.port = port
        self.url = url
        self.name = name
        self.default_args = {}
        self.log = logging.getLogger('Service.%s' % name)
        
        self.setDefaults(
            Service = 'AWSECommerceService',
            Version = '2007-10-29',
            AWSAccessKeyId = AWSAccessKeyId,
            Operation = 'ItemLookup',
            ItemPage = '1',
            ResponseGroup = 'Large',
            #ResponseGroup = 'Request,Small,ItemAttributes',
            Sort = 'salesrank',
        )
    
    def setDefaults(self, **kw_args):
        self.default_args.update(kw_args)
    
    def fetch(self, **kw_args):
        d = {}
        d.update(self.default_args)
        d.update(kw_args)
        
        d = self.prepare(d)
        
        query = urlencode(d)
        url = '%s?%s' % (self.url, query)
        self.log.debug('GET %s' % url)
        
        conn = HTTPConnection(self.server, self.port)
        conn.request('GET', url)
        r = conn.getresponse()
        data = None
        if r.status == httplib.OK:
            data = r.read()
        self.log.debug('RESPONSE %d %s' % (r.status, r.reason))
        return r.status, r.reason, data.decode('utf-8')
    
    def prepare(self, d):
        if d['IdType'] != 'ASIN':
            key = d['ItemId']
            key = self.NO_DIGITS.sub('', key)
            d['ItemId'] = key
        
        return d

def download(url):
    log = logging.getLogger('download')
    data = None
    log.debug('Downloading %s' % url)
    progress(tr('Downloading image...'))
    fh = urllib2.urlopen(url)
    data = fh.read()
    fh.close()
    # Need a buffer here because the image contains binary data
    # and sqlite3 won't accept it otherwise
    return buffer(data)

class TestService(Service):
    def __init__(self, name, results):
        self.name = name
        self.results = results
        self.log = logging.getLogger('Service.%s' % name)
        self.default_args = {}
    
    def fetch(self, **kw_args):
        d = {}
        d.update(self.default_args)
        d.update(kw_args)
        d = self.prepare(d)
        key = d['ItemId']
        self.log.debug('GET %s' % key)
        data = self.results.get(key)
        if data is None:
            return httplib.NOT_FOUND, 'Unknown key %s in %s (%s)' % (key, self.name, repr(d)), None
        
        return httplib.OK, 'OK', data

class Fetcher(object):
    def __init__(self, *services):
        self.services = services
        self.lastService = self.services[0]
        self.log = logging.getLogger('Fetcher')
        
    def fetch(self, key):
        if not self._fetch(self.lastService, key):
            nextService = None
            for service in self.services:
                # That didn't work a few seconds ago ... no need to try it again
                if service == self.lastService:
                    continue
                
                if self._fetch(service, key):
                    nextService = service
                    break
            
            if nextService is None:
                raise Exception('No service returned data for %s' % key)
            
            self.lastService = nextService
            
        msg = tr('Found data in Service "%1"', self.lastService.name)
        self.log.debug(msg)
        progress(msg)
        return self.root, self.response
        
    def _fetch(self, service, key):
        progress('Searching %s in "%s"' % (key, service.name))
        status, reason, data = service.fetch( 
            ItemId=key,
        )
        
        if status != httplib.OK:
            raise Exception('Request failed: %d %s' % (status, reason)) 
        
        #print repr(data)[:80]
        self.response = data
        # Save a copy of the request in a file for debugging purposes
        with codecs.open('last-request.xml', 'wb', 'utf-8') as fh:
            fh.write(self.response)
        
        builder = XMLTreeBuilder()
        builder.feed(data.encode('utf-8'))
        root = builder.close()
        
        self.root = XMLWrapper(root)

        Items = self.root.Items
        Request = Items.Request
        Errors = Request.Errors
        if Errors:
            l = Errors.Error
            if isinstance(l, XMLWrapper):
                l = [l]
            for e in l:
                msg = u'Service %s: %s' % (unicode(service.name), unicode(e.Message._text))
                progress(msg)
                self.log.warning(msg)
            return False
        
        return True

def progress(msg):
    pass

def stripNS(elem):
    tag = elem.tag
    if tag[0] == '{':
        return tag[1:].split("}", 1)[1]

    return tag

class AttributeWrapper(object):
    def __init__(self, elem):
        self._elem = elem

    def __getattr__(self, name):
        return self._elem.get(name)

class XMLWrapper(object):
    """A simple wrapper around ElementTree objects.
    
    Children will become Python attributes, you can
    access the XML attributes of the element via the
    python attribute '_'.
    """
    
    def __init__(self, elem, parent=None):
        #print 'wrapping',elem
        self._elem = elem
        self._name = stripNS(elem)
        self._parent = parent
        self._aWrapper = AttributeWrapper(elem)
        self._children = None
    
    def __getattr__(self, name):
        """Map XML to Python attributes.
        
        Use '_' to access the XML attributes of this node.
        
        Use '_text' to get the text stored in this node.
        
        Use the name of a child to access the child. You will
        get None if no child with this name exists, the child
        itself when there is only one child with this name or
        a list of children if there are several children with
        this name.
        """
        if name == '_':
            return self._aWrapper
        if name == '_text':
            return unicode(self._elem.text)
        
        #print 'Searching %s' % name
        # Doesn't work; the result is a weird list that changes its length and its content
        #result = self._elem.find('./%s' % name)
        if self._children is None:
            self._children = []
            for c in list(self._elem):
                self._children.append(XMLWrapper(c, self))
        
        result = []
        for c in self._children:
            if c._name == name:
                result.append(c)
        
        if len(result) == 1:
            return result[0]
        if len(result) == 0:
            return None
        return result
        
    def __repr__(self):
        return self._name

class Article(object):
    """A really primitive DB object.
    """
    TABLE_NAME = 'article'
    ALL_COLUMNS = (
        'searchIndex',
        'searchType',
        'searchKey',
        'asin',
        'detailurl',
        'type',
        'title',
        'creator',
        'image',
        'response',
        'location',
        'comment',
        'lentTo',
        'issue',
    )

    def __init__(self):
        self.id = None
        self.searchIndex = None
        self.searchType = None
        self.searchKey = None
        self.asin = None
        self.detailurl = None
        self.type = None
        self.title = None
        self.creator = None
        self.image = None
        self.response = None
        self.location = None
        self.comment = None
        self.lentTo = None
        self.issue = None
    
    def insert(self, cursor):
        """Insert the article into the DB
        """
        if not self.asin:
            raise Exception('missing asin')
        if not self.type:
            raise Exception('missing type')
        if not self.title:
            raise Exception('missing title')
        #if not self.searchIndex:
        #    raise Exception('missing searchIndex')
        if not self.searchType:
            raise Exception('missing searchType')
        if not self.searchKey:
            raise Exception('missing searchKey')
        
        #t = (self.searchIndex, self.searchType, self.searchKey, self.asin, self.detailurl, self.type, self.title, self.creator, self.image, self.response, self.location, self.comment)
        #for i in range(len(t)):
        #    print i,repr(t[i])[:80]
        #cursor.execute('''insert into article (searchIndex, searchType, searchKey, asin, detailurl, type, title, creator, image, response, location, comment) 
        #    values (?, ?, ?,  ?, ?, ?,  ?, ?, ?,  ?, ?, ?)''',
        #    t
        #)
        insertObject(cursor, self, Article.ALL_COLUMNS)
        
        rows = cursor.executeAndWait(DBRequest('''select id from article where asin = ?''', (self.asin,)))
        self.id = rows[0][0]
    
    def delete(self, db):
        db.execute(DBRequest('''delete from article where id = ?''', (self.id)))
        self.id = None
    
    def exists(self, cursor):
        """Returns True if this article is already in the DB.
        
        This is not 100% safe. The ASIN is unique but 
        there is a slight chance that two products have
        the same searchKey. Only, the chance that you
        actually own two products with the same searchKey
        is pretty small.
        """
        cursor.execute('select Count(*) from article where asin = ? or searchKey = ?', (self.asin, self.searchKey))
        return cursor.fetchone()[0] > 0
    
    def __repr__(self):
        return 'Article(search=(%s, %s, %s), asin=%s, type=%s, loc=%s, title=%s)' % (
            self.searchIndex, self.searchType, self.searchKey,
            self.asin,
            self.type,
            repr(self.location),
            repr(self.title),
        )

class Location(object):
    TABLE_NAME = 'location'
    ALL_COLUMNS = ('id', 'name', 'barcode')
    
    def __init__(self):
        self.id = None
        self.name = None
        self.barcode = None
    
    def insert(self, cursor):
        if not self.name:
            raise Exception('missing name')
        if not self.barcode:
            raise Exception('missing barcode')
        
        insertObject(cursor, self, Location.ALL_COLUMNS)
        #cursor.execute('''insert into %s (id, name, barcode) values (?,?,?)''', 
        #    (self.id, self.name, self.barcode))
        
        if self.id is None:
            rows = cursor.executeAndWait(DBRequest('''select id from location where barcode = ?''', (self.barcode,)))
            self.id = rows[0][0]
    
    def __repr__(self):
        return u'Location(id=%s, name=%s, barcode=%s' % (repr(self.id), self.name, self.barcode)

class Series(object):
    TABLE_NAME = 'series'
    ALL_COLUMNS = ('id', 'name', 'description')
    
    def __init__(self):
        self.id = None
        self.name = None
        self.description = None
        self.issues = {}
    
    def __repr__(self):
        return u'Series(id=%s, series=%s, name=%s, desc=%s)' % (
            repr(self.id), self.series, self.name, self.description
        )

def loadSeries(db, id_):
    loader = ObjectLoader(Series, Series.ALL_COLUMNS, id=id_)
    db.executeAndWait(loader)
    return loader.getSingleResult()

class Issue(object):
    TABLE_NAME = 'issue'
    ID = 'id'
    ASIN = 'asin'
    SERIES = 'series'
    NUMBER = 'number'
    NAME = 'name'
    DESCRIPTION = 'description'
    ALL_COLUMNS = (ID, ASIN, SERIES, NUMBER, NAME, DESCRIPTION)
    
    def __init__(self):
        self.id = None
        self.series = None
        self.number = 0
        self.name = None
        self.description = None
    
    def __repr__(self):
        return u'Issue(id=%s, series=%s, no=%d, name=%s, desc=%s)' % (
            repr(self.id), self.series, self.number, self.name, self.description
        )

def findIssue(db, asin):
    print 'findIssue asin:',asin
    loader = ObjectLoader(Issue, Issue.ALL_COLUMNS, asin=asin)
    db.executeAndWait(loader)
    return loader.getSingleResult()

def loadIssue(db, id_):
    loader = ObjectLoader(Issue, Issue.ALL_COLUMNS, id=id_)
    db.executeAndWait(loader)
    return loader.getSingleResult()

def createLocation(name, i):
    barcode = '20010001%04d' % i
    bc = EanBarCode()
    barcode = '%s%d' % (barcode, bc.computeChecksum(barcode))
    loc = Location()
    loc.name = name
    loc.barcode = barcode
    return loc

def insertObject(db, obj, columns):
    sql = 'insert into %s (%s) values (%s)' % (
        obj.TABLE_NAME, ', '.join(columns), ', '.join(['?'] * len(columns))
    )
    args = []
    for attr in columns:
        v = getattr(obj, attr)
        args.append(v)
    #log.debug('Execute SQL: %s %s' % (sql, args))
    req = DBRequest(sql, args)
    db.executeAndWait(req)

class DBRequest(object):
    def __init__(self, sql, args=None):
        self.sql = sql
        self.args = args
        # Set this event to cancel the request before it is execute.
        # It's safe to set this event after the request has been executed but
        # it won't help, either.
        self.cancel = Event()
        # Wait on this event to see when the request has been executed.
        self.done = Event()
        # If the request returns rows, you'll find them here
        self.result = None
        # This will be called with self.result as argument after the request has been processed
        self.callback = None
        # If this is not None, it will be called for each row fetched from
        # the database. The result of this call will be added to self.result
        # unless it's None.
        self.foreach = None
        # If an exception was thrown while processing this request, you'll find
        # it's value (see sys.exc_info()) here. This is usually the exception
        # message.
        self.error = None
    
    def setDone(self):
        self.done.set()

class CustomObjectLoader(DBRequest):
    def __init__(self, klass, fields, sql, *args):
        #self.log = logging.getLogger('ObjectLoader')

        self.klass = klass
        self.fields = fields
        
        super(CustomObjectLoader, self).__init__(sql, args)
    
    def setDone(self):
        result = []
        for row in self.result:
            obj = self.klass()
            
            for i in range(len(self.fields)):
                setattr(obj, self.fields[i], row[i])
            
            result.append(obj)
        
        self.result = result
        super(CustomObjectLoader, self).setDone()
    
    def getSingleResult(self):
        if len(self.result) == 0:
            return None
        if len(self.result) == 1:
            return self.result[0]
        
        raise Exception('Expected 1 result but got %d:\n%s' % (
            len(self.result), 
            '\n'.join ([repr(x) for x in self.result])
        ))
    
class ObjectLoader(CustomObjectLoader):
    def __init__(self, klass, fields, **where_cond):
        #self.log = logging.getLogger('ObjectLoader')

        sql = '''select %s from %s''' % (', '.join(fields), klass.TABLE_NAME)
        args = []
        
        if len(where_cond) > 0:
            l = []
            for key, value in where_cond.items():
                l.append('%s = ?' % key)
                args.append(value)
            
            sql = '%s where %s' % (sql, ' and '.join(l))
        
        super(ObjectLoader, self).__init__(klass, fields, sql, *args)
    
class DBUpdateRequest(DBRequest):
    def __init__(self, obj, fields):
        self.obj = obj
        self.fields = fields
        
        sql = 'update %s set %s where id = ?' % (
            obj.TABLE_NAME,
            ', '.join([('%s = ?' % f) for f in fields])
        )
        
        args = []
        for attr in fields:
            v = getattr(obj, attr)
            args.append(v)
        args.append(obj.id)
        
        super(DBUpdateRequest, self).__init__(sql, args)

class DBExportRequest(DBRequest):
    def __init__(self, sql, args, fh, tag, fields):
        super(DBExportRequest, self).__init__(sql, args)
        self.fh = fh
        self.tag = tag
        self.fields = fields

        self.foreach = self.exportRow

    def exportRow(self, row):
        self.fh.write('\t\t<%s>\n' % self.tag)
        
        for i in range(len(self.fields)):
            field = self.fields[i]
            value = row[i]
            
            if isinstance(value, types.BufferType):
                self.fh.write('\t\t\t<%s>\n' % field)
                self.fh.write(base64.encodestring(value[:]))
                self.fh.write('\t\t\t</%s>\n' % field)
                continue
            
            if not isinstance(value, types.StringTypes):
                value = unicode(value)
            
            self.fh.write('\t\t\t<%s>' % field)
            if '<' in value or '>' in value or '&' in value:
                if len(value) > 64:
                    self.fh.write('<![CDATA[')
                    self.fh.write(value)
                    self.fh.write(']]>')
                else:
                    value = value.replace('<', '&lt;')
                    value = value.replace('>', '&gt;')
                    value = value.replace('&', '&amp;')
                    self.fh.write(value)
            else:
                self.fh.write(value)
            self.fh.write('</%s>\n' % field)

        self.fh.write('\t\t</%s>\n' % self.tag)
        return None

class DB(Thread):
    def __init__(self, dbName, isolation_level=None):
        super(DB, self).__init__(name='DB Thread')
        self.log = logging.getLogger('DB')
        self.dbName = dbName
        self.db = None
        self.cursor = None
        self.isolation_level = isolation_level
        self.requests = Queue()
    
    def backup(self):
        if self.dbName == ':memory:':
            return
        
        if self.isConnected():
            raise Exception("Can't backup a live DB")
        
        import shutil
        import time
        
        destName = '%s-%s.bak' % (dbName, time.strftime('%Y%m%d-%H%M%S'))
        shutil.copyfile(path, destName)
    
    def isConnected(self):
        return self.db is not None
    
    SHUTDOWN = 'SHUTDOWN'
    def close(self):
        if self.isConnected() or self.isAlive():
            self.log.debug('Shutdown')
            req = DBRequest(DB.SHUTDOWN)
            self.execute(req)
            self.log.debug('Waiting for thread to die')
            req.done.wait(10)
            self.log.debug('DB successfully stopped')
        
    def execute(self, request):
        """Execute a DBRequest."""
        self.log.debug('Adding request to queue: %s' % request.sql)
        self.requests.put(request)
    
    def executeAndWait(self, request):
        self.requests.put(request)
        request.done.wait()
        return request.result
    
    def run(self):
        self.log.debug('Waiting for work')
        try:
            while True:
                request = self.requests.get()
                log.debug('Got %s' % repr(request.sql))
                if request.sql == DB.SHUTDOWN:
                    request.done.set()
                    break
                
                if request.cancel.isSet():
                    log.debug('Request is canceled')
                    continue
                
                if self.db is None:
                    self.db = sqlite3.connect(self.dbName, isolation_level=self.isolation_level)
                    self.cursor = self.db.cursor()
                
                args = request.args
                if args is None:
                    args = ()
                elif not isinstance(args, (types.ListType, types.TupleType)):
                    args = (args,)
                try:
                    log.debug(('Execute SQL: %s %s' % (request.sql, args))[:100])
                    self.cursor.execute(request.sql, args)
                    
                    if request.foreach is None:
                        request.result = list(self.cursor.fetchall())
                    else:
                        request.result = []
                        while True:
                            row = self.cursor.fetchone()
                            if row is None:
                                break
                            
                            try:
                                row = request.foreach(row)
                            except:
                                log.exception('request.foreach failed for: %s' % (repr(row),))
                                value = sys.exc_info()[1]
                                request.error = value
                                break
                            if row is not None:
                                request.result.append(row)
                    log.debug('Got %d results' % (len(request.result),))
                except:
                    log.exception('Exception while executing SQL: %s %s' % (repr(request.sql), args))
                    value = sys.exc_info()[1]
                    request.error = value
                
                request.setDone()
                if request.callback is not None:
                    try:
                        request.callback(request.result)
                    except:
                        log.exception('Exception in callback %s of %s' % (request.callback, request.sql))
            
        finally:
            try:
                if self.cursor is not None:
                    self.cursor.close()
                    self.cursor = None
            except:
                self.log.exception('Error closing cursor')
        
            try:
                if self.db is not None:
                    self.db.close()
                    self.db = None
            except:
                self.log.exception('Error closing db')
            
        self.log.debug('DB thread stopped')


def createArticle(root, response):
    """Create an article from the Amazon response."""
    log = logging.getLogger('createArticle')
    #print repr(root._name)
    OperationRequest = root.OperationRequest
    RequestId = OperationRequest.RequestId
    #print RequestId
    #print repr(RequestId._text)
    if 0 == 1:
        Arguments = OperationRequest.Arguments
        SearchIndex = None
        ItemId = None
        IdType = None
        for Argument in Arguments.Argument:
            a = Argument._
            Name = a.Name
            Value = a.Value
            
            if Name == 'SearchIndex':
                SearchIndex = Value
            elif Name == 'ItemId':
                ItemId = Value
            elif IdType == 'IdType':
                IdType = Value
        print SearchIndex, IdType, ItemId

    Items = root.Items
    Request = Items.Request
    if Request.IsValid._text != 'True':
        raise Exception('Request failed!')

    ItemLookupRequest = Request.ItemLookupRequest
    a = Article()
    a.searchIndex = ItemLookupRequest.SearchIndex
    if a.searchIndex:
        a.searchIndex = a.searchIndex._text
    a.searchType = ItemLookupRequest.IdType._text
    a.searchKey = ItemLookupRequest.ItemId._text
    #print ItemLookupRequest.SearchIndex._text, ItemLookupRequest.IdType._text, ItemLookupRequest.ItemId._text
    #print a

    Item = Items.Item
    if type(Item) == type([]):
        raise Exception('Found more than one item!')

    a.asin = Item.ASIN._text
    #print a.asin
    a.detailurl = Item.DetailPageURL._text
    pos = a.detailurl.find('%3F')
    if pos == -1:
        pos = a.detailurl.find('?')
    if pos != -1:
        a.detailurl = a.detailurl[:pos]
    
    #print a.detailurl
    data = None
    if Item.LargeImage:
        imageURL = Item.LargeImage.URL._text
        #print imageURL
        data = download(imageURL)
        ext = imageURL[-4:]
        #with codecs.open('last-image%s' % ext, 'wb', ISO_8859_1) as fh:
        # Keep a backup of the image for debugging purposes
        with open('last-image%s' % ext, 'wb') as fh:
            fh.write(data)
    else:
        log.warning('No image found')    
    a.image = data
    a.response = response
    
    creator = None
    a.type = Item.ItemAttributes.ProductGroup._text
    if a.type == 'Book':
        creator = Item.ItemAttributes.Author
        a.type = 'Buch'
    elif a.type == 'DVD':
        creator = Item.ItemAttributes.Author
        if creator is None:
            creator = Item.ItemAttributes.Creator
    elif a.type == 'Music':
        creator = Item.ItemAttributes.Author
        if creator is None:
            creator = Item.ItemAttributes.Creator
        if creator is None:
            creator = Item.ItemAttributes.Artist
        
    if creator is None:
        log.warning('No author/creator/artists found!')
    elif type(creator) == type([]):
        l = []
        for c in creator:
            l.append(c._text)
        creator = ', '.join(l)
    else:
        creator = creator._text
    a.creator = creator
    a.title = Item.ItemAttributes.Title._text
    
    return a

def createDB(db):
    log = logging.getLogger('createDB')
    
    log.info('Creating tables')
    db.execute(DBRequest('''
    create table article(
        id INTEGER PRIMARY KEY AUTOINCREMENT,
        searchIndex TEXT,
        searchType TEXT,
        searchKey TEXT,
        asin TEXT,
        detailurl TEXT,
        type TEXT,
        title TEXT,
        creator TEXT,
        image BLOB,
        response TEXT,
        location INTEGER,
        comment TEXT,
        lentTo INTEGER,
        issue INTEGER,
        UNIQUE(asin)
    )
    '''))
    db.execute(DBRequest('''
    create table location(
        id INTEGER PRIMARY KEY AUTOINCREMENT,
        name TEXT,
        barcode TEXT,
        UNIQUE(barcode)
    )
    '''))
    db.execute(DBRequest('''
    create table addinfo(
        id INTEGER PRIMARY KEY AUTOINCREMENT,
        key TEXT,
        type TEXT,
        value TEXT
    )
    '''))
    db.execute(DBRequest('''
    create table series(
        id INTEGER PRIMARY KEY AUTOINCREMENT,
        name TEXT NOT NULL,
        description TEXT,
    )
    '''))
    db.execute(DBRequest('''
    create table issue(
        id INTEGER PRIMARY KEY AUTOINCREMENT,
        asin TEXT,
        series INTEGER NOT NULL,
        number INTEGER NOT NULL,
        name TEXT NOT NULL,
        description TEXT,
    )
    '''))
    
    i = 1
    for name in ('Home', 'Somewhere else', 'At work', 'Basement', 'Roof',
        'Lost', 'DVD Box 1', 'DVD Box 2', 'DVD Box 3', 'DVD Box 4',
        'DVD Box 5', 'DVD Box 6', 'DVD Box 7', 'DVD Box 8', 
        'Tommy', 'Mother', 'Wishlist',):
        loc = createLocation(name, i)
        i += 1
        loc.insert(db)
        
    db.execute(DBRequest('''
    create table version(
        version TEXT
    )
    '''))
    db.execute(DBRequest('''insert into version values (2)'''))
    
def upgradeDB(path):
    log = logging.getLogger('upgradeDB')
    log.info('Connecting to DB %s' % repr(path))
    
    if path == ':memory:' or not os.path.exists(path):
        return
    
    import shutil
    import time
    
    destName = '%s-%s.bak' % (dbName, time.strftime('%Y%m%d-%H%M%S'))
    shutil.copyfile(path, destName)
    
    db = sqlite3.connect(path, isolation_level=None)
    c = db.cursor()
    
    #c.execute('''select id, detailurl from article where detailurl glob '*%3F*' ''')
    #rows = c.fetchall()
    #for row in rows:
    #    id_ = row[0]
    #    detailurl = row[1]
    #    pos = detailurl.find('%3F')
    #    if pos == -1:
    #        pos = detailurl.find('?')
    #    if pos == -1:
    #        continue
    #    detailurl = detailurl[:pos]
    #    print id_, detailurl
    #    c.execute('''update article set detailurl = ? where id = ?''', (detailurl, id_))
    
    row = None
    try:
        c.execute('select version from version')
        row = c.fetchone()
    except:
        log.exception("Reading version from DB failed")
    
    if row is None:
        version = 1
    else:
        version = int(row[0])
    
    log.debug('Found version %d' % version)
    if version == 1:
        log.info('Upgrading DB to version 2')
        c.execute('''
        create table version(
            version TEXT
        )
        ''')
        c.execute('''
        insert into version values (1)
        ''')
        
        c.execute('''
        alter table article
        add column comment
        ''')
        
        version = 2
        c.execute('update version set version = ?', (version,))
    
    if version == 2:
        log.info('Upgrading DB to version 3')
        c.execute('''
        create table addinfo(
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            key TEXT,
            type TEXT,
            value TEXT
        )
        ''')
        c.execute('''
        create table series(
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            name TEXT NOT NULL,
            description TEXT
        )
        ''')
        c.execute('''
        create table issue(
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            asin TEXT NOT NULL,
            series INTEGER NOT NULL,
            number INTEGER NOT NULL,
            name TEXT NOT NULL,
            description TEXT
        )
        ''')
        c.execute('''
        alter table article
        add column lentTo INTEGER
        ''')
        c.execute('''
        alter table article
        add column issue INTEGER
        ''')
        
        version = 3    
        c.execute('update version set version = ?', (version,))
    
    c.close()
    db.close()
    
def testDownload(url):
    with open('test-image.jpg', 'rb') as fh:
        data = fh.read()
    return buffer(data)

def prepareTestXML(filename, **kw_args):
    if os.path.exists(filename):
        with codecs.open(filename, 'rb', 'utf-8') as fh:
            xml = fh.read()
    else:
        print 'Downloading %s' % filename
        s = Service('webservices.amazon.de', url='/onca/xml')
        x,y,xml = s.fetch(**kw_args)
        with codecs.open(filename, 'wb', 'utf-8') as fh:
            fh.write(xml)
    return xml

def runTests():
    global download
    madagascar_xml = prepareTestXML('madagascar.xml',
            IdType='EAN',
            SearchIndex='DVD',
            ItemId='4047552500073',
    )
    hirnforschung_xml = prepareTestXML('hirnforschung.xml',
            IdType='EAN',
            SearchIndex='Books',
            ItemId='9783518260043',
    )
    no_data_xml = prepareTestXML('no-data.xml',
            IdType='EAN',
            SearchIndex='DVD',
            ItemId='X4047552500073',
    )
    music_xml = prepareTestXML('music-test.xml',
            IdType='UPC',
            SearchIndex='Music',
            ItemId='743212484120',
    )
    
    if not os.path.exists('test-image.jpg'):
        print 'Downloading test image...'
        data = download('http://ecx.images-amazon.com/images/I/519G%2B8Wcl8L._SL500_.jpg')
        with open('test-image.jpg', 'wb') as fh:
            fh.write(data)
    
    # Add a few umlauts to the title to make sure the Python unicode mess works... 
    hirnforschung_xml = hirnforschung_xml.replace(u'Hirnforschung', u'Hirnfrschng')
    
    def not_found(**kw_args):
        s = no_data_xml
        s = s.replace(u'<IdType>EAN</IdType>', u'<IdType>%s</IdType>' % kw_args['IdType'])
        s = s.replace(u'X4047552500073', unicode(kw_args['ItemId']))
        s = s.replace(u'<SearchIndex>DVD</SearchIndex>', u'<SearchIndex>%s</SearchIndex>' % kw_args['SearchIndex'])
        return s
    
    books = TestService('books', {
        '9783518260043': hirnforschung_xml,
        '4047552500073': not_found(ItemId='4047552500073', SearchIndex='Books', IdType='EAN',),
        'X4047552500073': not_found(ItemId='X4047552500073', SearchIndex='Books', IdType='EAN',),
        '743212484120': not_found(ItemId='743212484120', SearchIndex='Books', IdType='EAN',),
    })
    books.setDefaults(
        IdType='EAN',
        SearchIndex='Books',
    )
    dvds = TestService('dvds', {
        '4047552500073': madagascar_xml,
        'X4047552500073': not_found(ItemId='X4047552500073', SearchIndex='DVD', IdType='EAN',),
        '9783518260043': not_found(ItemId='9783518260043', SearchIndex='DVD', IdType='EAN',),
        '743212484120': not_found(ItemId='743212484120', SearchIndex='DVD', IdType='EAN',),
    })
    dvds.setDefaults(
        IdType='EAN',
        SearchIndex='DVD',
    )
    music = TestService('music', {
        '743212484120': music_xml,
        '9783518260043': not_found(ItemId='9783518260043', SearchIndex='Music', IdType='UPC',),
    })
    music.setDefaults(
        IdType='UPC',
        SearchIndex='Music',
    )
    fetcher = Fetcher(books, dvds, music)

    testTr()

    download = testDownload
    db = DB(':memory:')
    try:
        db.start()
        createDB(db)
        
        testNotFound(fetcher)
        testBook(fetcher, db)
        testDVD(fetcher, db)
        testMusic(fetcher, db)
    finally:
        db.close()
    
    print
    print 'All tests passed.'

def testTr():
    s = tr('a')
    assert 'a' == s, s
    
    s = tr('%b')
    assert '%b' == s, s
    s = tr('a%b')
    assert 'a%b' == s, s
    s = tr('a%')
    assert 'a%' == s, s
    
    s = tr('%1', 'b')
    assert 'b' == s, s
    s = tr('a%1', 'b')
    assert 'ab' == s, s
    s = tr('a%1c', 'b')
    assert 'abc' == s, s
    s = tr('%1c', 'b')
    assert 'bc' == s, s
    
    from PyQt4.QtCore import QString
    s = unicode(QString(u"%%a%1b%2c%3d").arg(*(u'1',u'%2',u'3')))
    assert '%%a1b%2c3d' == s, s
    
    s = tr(u"%%a%1b%2c%3d",u'1',u'%2',u'3')
    assert '%%a1b%2c3d' == s, s
    
def testDVD(fetcher, db):
    print 'testDVD'
    root, response = fetcher.fetch('4047552500073')
    assert fetcher.lastService.name == 'dvds'
    a = createArticle(root, response)
    expected = 'Article(search=(DVD, EAN, 4047552500073), asin=B000FO43F4, type=DVD, loc=None, title=u\'Madagascar\')'
    assert repr(a) == expected, '\n%s\n%s' % (repr(a), expected)
    
    a.insert(db)
    assert a.id == 2, a.id
    
    loader = ObjectLoader(Article, Article.ALL_COLUMNS, id=a.id)
    db.executeAndWait(loader)
    b = loader.result[0]
    
    assert b.detailurl == a.detailurl

def testBook(fetcher, db):
    print 'testBook'
    root, response = fetcher.fetch('9783518260043')
    assert fetcher.lastService.name == 'books'
    a = createArticle(root, response)
    expected = 'Article(search=(Books, EAN, 9783518260043), asin=3518260049, type=Buch, loc=None, title=u\'Hirnf\\xf6rsch\\xfcng und Meditation: Ein Dialog\')'
    #print repr(repr(a.title))
    assert repr(a) == expected, '\n%s\n%s' % (repr(a), expected)
    
    a.insert(db)
    assert a.id == 1, a.id
    
    loader = ObjectLoader(Article, Article.ALL_COLUMNS, id=a.id)
    db.executeAndWait(loader)
    b = loader.result[0]
    
    assert a.image == b.image, '\n%s\n%s' % (repr(a.image)[:80], repr(b.image)[:80])
    assert a.detailurl == b.detailurl
    assert a.response == b.response

def testMusic(fetcher, db):
    print 'testMusic'
    root, response = fetcher.fetch('7 4321-24841-2 0')
    assert fetcher.lastService.name == 'music'
    a = createArticle(root, response)
    expected = 'Article(search=(Music, UPC, 743212484120), asin=B000025ATD, type=Music, loc=None, title=u\'Danzig\')'
    #print repr(repr(a.title))
    assert repr(a) == expected, '\n%s\n%s' % (repr(a), expected)
    
    a.insert(db)
    assert a.id == 3, a.id
    
    loader = ObjectLoader(Article, Article.ALL_COLUMNS, id=a.id)
    db.executeAndWait(loader)
    b = loader.result[0]
    
    assert a.image == b.image, '\n%s\n%s' % (repr(a.image)[:80], repr(b.image)[:80])
    assert a.detailurl == b.detailurl
    assert a.response == b.response

def testNotFound(fetcher):
    print 'testNotFound'
    try:
        fetcher.fetch('X4047552500073')
        raise Exception('Test failed: no exception was thrown')
    except:
        pass

def switchLocation(c, key, location):
    c.execute('select id, name from location where barcode = ?', (key,))
    row = c.fetchone()
    if row is None:
        log.info('No location defined for barcode %s' % key)
        print 'ID or name for location or RETURN to abort'
        s = raw_input('> ')
        _id = None
        try:
            _id = int(s)
        except:
            pass
        
        if s.strip() == '':
            log.info('Switching location aborted.')
            return location
        elif _id is None:
            log.info('Creating new location "%s"' % s)
            c.execute('insert into location (name, barcode) values (?,?)', (s, key))
            c.execute('select id, name from location where barcode = ?', (key,))
            row = c.fetchone()
        else:
            while True:
                c.execute('select id, name from location where id = ?', (_id,))
                row = c.fetchone()
                if row is not None:
                    c.execute('update location set barcode = ? where id = ?', (key, _id))
                    log.info('Barcode %s assigned to location "%s"' % (key, row[1]))
                    break
    
    log.info('New location selected: %d %s' % (row[0], row[1]))
    return row[0]

def setupFetcher():
    books_amazon_de = Service('webservices.amazon.de', url='/onca/xml', name=u'Deutsche Bcher')
    books_amazon_de.setDefaults(
        IdType='EAN',
        SearchIndex='Books',
    )
    dvd_amazon_de = Service('webservices.amazon.de', url='/onca/xml', name=u'Deutsche DVD')
    dvd_amazon_de.setDefaults(
        IdType='EAN',
        SearchIndex='DVD',
    )
    music_amazon_de = Service('webservices.amazon.de', url='/onca/xml', name=u'Deutsche Musik CDs')
    music_amazon_de.setDefaults(
        IdType='UPC',
        SearchIndex='Music',
    )
    video_games_amazon_de = Service('webservices.amazon.de', url='/onca/xml', name=u'Deutsche DVD')
    video_games_amazon_de.setDefaults(
        IdType='EAN',
        SearchIndex='VideoGames',
    )
    books_amazon_com = Service('webservices.amazon.com', url='/onca/xml', name=u'English Book')
    books_amazon_com.setDefaults(
        IdType='EAN',
        SearchIndex='Books',
    )
    dvd_amazon_com = Service('webservices.amazon.com', url='/onca/xml', name=u'English DVD')
    dvd_amazon_com.setDefaults(
        IdType='EAN',
        SearchIndex='DVD',
    )
    music_amazon_com = Service('webservices.amazon.com', url='/onca/xml', name=u'English Music CDs')
    music_amazon_com.setDefaults(
        IdType='UPC',
        SearchIndex='Music',
    )
    asin_amazon_de = Service('webservices.amazon.de', url='/onca/xml', name=u'Deutsche DVD')
    asin_amazon_de.setDefaults(
        IdType='ASIN',
    )
    asin_amazon_com = Service('webservices.amazon.com', url='/onca/xml', name=u'Deutsche DVD')
    asin_amazon_com.setDefaults(
        IdType='ASIN',
    )
    
    fetcher = Fetcher(
        books_amazon_de,
        dvd_amazon_de,
        # Doesn't work; can't search EANs of PS2/3 games
        #video_games_amazon_de,
        books_amazon_com,
        dvd_amazon_com,
        music_amazon_de,
        music_amazon_com,
        asin_amazon_de,
        asin_amazon_com,
    )
    
    return fetcher
    
def run(db):
    log = logging.getLogger('run')
    
    #status, reason, data = amazon_de.fetch( 
    #    ItemId='9783518260043',
    #)
    #if status != httplib.OK:
    #    raise 'Request failed: %d %s' % (status, reason) 
    # 
    #print repr(data)    
    
    c = db.cursor()
    c.execute('select Count(*) from article')
    log.info('Anzahl Artikel: %d' % c.fetchone()[0])
    
    log.info('Welcome to UPCScan %s.' % VERSION)
    
    location = 0
    c.execute('select id, name from location where id = ?', (location,))
    row = c.fetchone()
    if row is not None:
        log.info('Current location: "%s"', row[1])
    else:
        log.info('No Current location!')

    # Use these to create test data
    #root, response = fetcher.fetch('9783518260043')
    #root, response = fetcher.fetch('4047552500073')
    #root, response = fetcher.fetch('X4047552500073')
    if len(sys.argv) == 1:
        class Source(object):
            def next(self):
                return sys.stdin.readline().strip()
        source = Source()
        log.info('Reading codes from stdin. Press RETURN (without a code) or Ctrl-D to exit.')
    else:
        class Source(object):
            def __init__(self, args):
                self.args = args
                self.i = 0
                
            def next(self):
                if self.i < len(self.args):
                    i = self.i
                    self.i += 1
                    return self.args[i]
                else:
                    return None
        source = Source(sys.argv[1:])
    
    fetcher = setupFetcher()
    
    while True:
        key = source.next()
        if key is None or not key.strip():
            break
        
        # Internal key
        if key[:3] == '200':
            location = switchLocation(c, key, location)
            continue
        
        a = Article()
        a.searchKey = key
        a.asin = None
        if a.exists(c):
            a.load(c, searchKey=key)
            log.warning('Article already exists: %s' % repr(a))
            continue
        
        try:
            root, response = fetcher.fetch(key)
        except Exception, e:
            log.warning('Article %s not found anywhere. Adding it to "failed-codes"' % key)
            with open('failed-codes', 'a') as fh:
                fh.write('%s %s\n' % (key, repr(e)))
            continue
        
        a = createArticle(root, response)
        if a.exists(c):
            log.warning('Article already exists: %s' % repr(a))
            continue
        
        a.location = location
        print a
        
        issue = findIssue(db, a.asin)
        if issue is not None:
            a.issue = issue.id
        
        a.insert(c)
        #print a.id
        
        #b = Article()
        #b.load(c, a.id)
        #print b
        #c.commit()
    
    c.execute('select Count(*) from article')
    log.info('Anzahl Artikel: %d' % c.fetchone()[0])
    c.close()
    log.info('Done.')

def runQt4(db):
    from PyQt4.QtGui import \
        QMainWindow, QItemSelection, QItemSelectionModel, \
        QLabel, QLineEdit, QTableView, QApplication, QWidget, \
        QVBoxLayout, QHBoxLayout, QAbstractItemView, \
        QGridLayout, QPushButton, QImageReader, QIcon, QPixmap, QImage, \
        QDialog, QStatusBar, QColor, QSortFilterProxyModel, QComboBox, \
        QFileDialog, QMessageBox
    from PyQt4.QtCore import \
        SIGNAL, QVariant, Qt, QModelIndex, QSize, QBuffer, QByteArray, \
        QString, QAbstractTableModel, QObject, SLOT, QCoreApplication, \
        QFile, QIODevice
    global tr, tr2
    def tr(*args):
        if len(args) == 1:
            return QCoreApplication.translate(u'upcscan', QString(args[0]))
        return QCoreApplication.translate(u'upcscan', QString(args[0]).arg(*args[1:]))
    def tr2(contex, *args):
        return QCoreApplication.translate(contex, QString(args[0]).arg(*args[1:]))
    
    fetcher = setupFetcher()

    def bufferToImage(imgData):
        """Convert a Python string or buffer into a QImage."""
        # Holy crap ... it's hard to convert a Python buffer into something
        # that Qt will accept.
        s = QByteArray(imgData[:])
        b = QBuffer(s)
        b.open(QBuffer.ReadOnly)
        reader = QImageReader(b)
        
        # Try to figure out the format
        reader.canRead()
        format = unicode(QString(reader.format()))
        #print 'format',repr(format)
        
        img = QImage()
        if not reader.read(img):
            log = logging.getLogger('bufferToImage')
            log.error('Error loading image from buffer (%d): %s' % (
                    len(imgData),
                    reader.errorString(),
                )
            )
            img = None
        b.close()
        return img, format
        
    def bufferToPixmap(imgData):
        img, format = bufferToImage(imgData)
        if img is None:
            return None
        return QPixmap.fromImage(img)

    class CoverDialog(QDialog):
        MODE_DOWNLOAD = 'download'
        MODE_ASSIGN = 'assign'
        
        def __init__(self, parent=None):
            super(CoverDialog, self).__init__(parent)
            self.log = logging.getLogger('CoverDialog')
            self.setWindowTitle(tr(u'Cover Image'))
            
            self.main = parent
            self.db = self.main.db
            
            self.mode = CoverDialog.MODE_DOWNLOAD
            
            self.createWidgets()
            
            self.setMode(CoverDialog.MODE_DOWNLOAD)
            self.imgDir = os.getcwdu()
        
        def createWidgets(self):
            vbox = QVBoxLayout()
            self.setLayout(vbox)
            grid = QGridLayout()
            vbox.addLayout(grid)
            
            row = 0
            self.coverImageWidget = QLabel()
            size = QSize(128, 128)
            self.coverImageWidget.resize(size)
            grid.addWidget(self.coverImageWidget, row, 0, 1, 3)
            row += 1
            
            label = QLabel(tr(u'&Update from:'))
            grid.addWidget(label, row, 0)
            self.urlWidget = QLineEdit()
            label.setBuddy(self.urlWidget)
            grid.addWidget(self.urlWidget, row, 1)
            self.fileButton = QPushButton()
            self.fileButton.setText(tr(u'&...'))
            self.connect(self.fileButton, SIGNAL('clicked(bool)'), self.fileButtonClicked)
            grid.addWidget(self.fileButton, row, 2)
            row += 1
            
            hbox = QHBoxLayout()
            vbox.addLayout(hbox)
            
            self.okButton = QPushButton()
            self.connect(self.okButton, SIGNAL(u'clicked(bool)'), self.okButtonClicked)
            hbox.addWidget(self.okButton)
            
            self.saveAsButton = QPushButton()
            self.saveAsButton.setText(tr(u'&Save as...'))
            self.connect(self.saveAsButton, SIGNAL('clicked(bool)'), self.saveAsButtonClicked)
            hbox.addWidget(self.saveAsButton)
            
            self.closeButton = QPushButton()
            self.closeButton.setText(tr(u'&Close'))
            self.connect(self.closeButton, SIGNAL('clicked(bool)'), self, SLOT('reject()'))
            hbox.addWidget(self.closeButton)
        
        def saveAsButtonClicked(self, checked):
            format = self.coverImageFormat
            if not format:
                format = 'png'
            if format.lower() == 'jpeg':
                format = 'jpg'
            filename = QFileDialog.getSaveFileName(self,
                tr(u'Save image as ...'),
                self.imgDir,
                tr(u'''Image (*.%1);;All Images (*.png *.gif *.jpg *.jpeg);;All files (*.*)''', format),
            )
            filename = unicode(filename)
            if filename:
                #print repr(filename)
                self.imgDir = os.path.dirname(filename)
                
                root, ext = os.path.splitext(filename)
                newFormat = ext[1:].lower()
                if newFormat.lower() == 'jpeg':
                    newFormat = 'jpg'
                
                self.log.debug('Old format: %s, new format: %s' % (format, newFormat))
                if newFormat == format:
                    # No format conversion
                    with open(filename, 'wb') as fh:
                        fh.write(self.coverData)
                else:
                    # Save the image in a new format
                    self.coverImage.save(filename, newFormat)
                progress(tr('Image saved as %1', filename))
        
        def okButtonClicked(self, checked):
            print self.main.currentArticle
            if self.mode == CoverDialog.MODE_DOWNLOAD:
                url = unicode(self.urlWidget.text())
                if not url.strip():
                    return
                
                try:
                    self.coverData = download(url)
                except:
                    log.exception('Download from %s failed' % url)
                    return
                
                progress(tr(u'Ready.'))
                self.coverImage = None
                self.coverPixmap = None
                self.coverImageFormat = ''
                if self.coverData is not None:
                    self.coverImage, self.coverImageFormat = bufferToImage(self.coverData)
                if self.coverImage is not None:
                    self.pixmap = QPixmap.fromImage(self.coverImage)
                self.updatePixmap()
                
                self.setMode(CoverDialog.MODE_ASSIGN)
            else:
                # Assign mode
                a = Article()
                a.id = self.main.currentArticle.id
                a.image = self.coverData
                req = DBUpdateRequest(a, ('image',))
                self.db.execute(req)
                
                self.main.coverData = self.coverData
                self.main.updateCoverWidget()
                self.setMode(CoverDialog.MODE_DOWNLOAD)

        def fileButtonClicked(self, checked):
            filename = QFileDialog.getOpenFileName(self,
                tr(u'Select Image File'),
                self.imgDir,
                tr(u'''Images (*.png *.gif *.jpg *.jpeg);;All files (*.*)'''),
            )
            filename = unicode(filename)
            if filename:
                print repr(filename)
                self.imgDir = os.path.dirname(filename)
                self.urlWidget.setText(filename)
                
                fh = QFile(filename)
                try:
                    fh.open(QIODevice.ReadOnly)
                    self.coverData = buffer(fh.readAll())
                finally:
                    fh.close()
                
                self.coverImage, self.coverImageFormat = bufferToImage(self.coverData)
                
                self.pixmap = QPixmap.fromImage(self.coverImage)
                self.updatePixmap()
                #self.currentArticle = currentArticle
                #print 'setCurrentArticle',self.currentArticle
                self.setMode(CoverDialog.MODE_ASSIGN)
        
        def setMode(self, mode):
            self.mode = mode
            if self.mode == CoverDialog.MODE_DOWNLOAD:
                self.okButton.setText(u'&Download')
            else:
                self.okButton.setText(tr(u'&Assign'))
            self.okButton.setEnabled(self.main.currentArticle is not None)
        
        def updateCoverWidget(self):
            self.pixmap = self.main.coverPixmap
            self.coverData = self.main.coverData
            self.coverImage = self.main.coverImage
            self.coverImageFormat = self.main.coverImageFormat
            self.updatePixmap()
        
        def updatePixmap(self):
            if self.pixmap is None:
                self.pixmap = QPixmap()
                self.coverImageWidget.setPixmap(self.pixmap)
                self.coverImageWidget.setText(tr(u'No cover image'))
                self.saveAsButton.setEnabled(False)
            else:
                size = self.pixmap.size()
                self.coverImageWidget.setPixmap(self.pixmap)
                self.coverImageWidget.resize(size)
                self.coverImageWidget.setMinimumSize(size)
                self.saveAsButton.setEnabled(True)
    
    class Column(object):
        def __init__(self, name, headerAlignment=(Qt.AlignLeft | Qt.AlignVCenter), 
                editable=False, dataAlignment=(Qt.AlignLeft | Qt.AlignVCenter),
                dataColor=QColor(Qt.black), backgroundColor=(QColor(Qt.white),)):
            self.name = QVariant(name)
            self.headerAlignment = QVariant(headerAlignment)
            self._dataAlignment = QVariant(dataAlignment)
            self.editable = editable
            self._dataColor = QVariant(dataColor)
            self._backgroundColor = [QVariant(c) for c in backgroundColor]
        
        def dataAlignment(self, obj):
            return self._dataAlignment
        
        def dataColor(self, obj):
            return self._dataColor
        
        def backgroundColor(self, index, obj):
            return self._backgroundColor[index.row() % len(self._backgroundColor)]
        
        def toString(self, obj):
            '''Convert the value in obj for this column into a unicode string (not a QVariant!).'''
            return tr(u'UNDEF')
        
    class PyTableModel(QAbstractTableModel):
        def __init__(self, parent=None):
            super(PyTableModel, self).__init__(parent)
            self.verticalTextAlignment = QVariant(Qt.AlignRight | Qt.AlignVCenter)
            self.header = []
            self.dataList = []
        
        def rowCount(self, index=QModelIndex()):
            return len(self.dataList)
        
        def columnCount(self, index=QModelIndex()):
            return len(self.header)
        
        def addColumn(self, c):
            self.header.append(c)
            return c
        
        def data(self, index, role=Qt.DisplayRole):
            if not index.isValid() or \
                not (0 <= index.row() <= self.rowCount()):
                return QVariant()
            
            column = index.column()
            obj = self.dataList[index.row()]
            if role == Qt.TextAlignmentRole:
                return self.header[column].dataAlignment(obj)
            elif role == Qt.TextColorRole:
                return self.header[column].dataColor(obj)
            elif role == Qt.BackgroundColorRole:
                return self.header[column].backgroundColor(index, obj)
            elif role == Qt.DisplayRole:
                return QVariant(self.header[column].toString(obj))
            
            return QVariant()
        
        def setData(self, index, value, role=Qt.EditRole):
            try:
                if index.isValid() and 0 <= index.row() < self.rowCount():
                    #print index,index.row(),index.column()
                    self.emit(SIGNAL('dataChanged(const QModelIndex&,const QModelIndex&)'), index, index)
                    return True
            except:
                log.exception('Error in setData')
            
            return False
        
        def headerData(self, section, orientation, role=Qt.DisplayRole):
            if orientation == Qt.Vertical:
                if role == Qt.TextAlignmentRole:
                    return self.verticalTextAlignment
                elif role == Qt.DisplayRole:
                    return QVariant(int(section + 1))
                
                return QVariant()
            
            header = self.header[int(section)]
            if role == Qt.TextAlignmentRole:
                return header.headerAlignment
            elif role == Qt.DisplayRole:
                return header.name
            
            return QVariant()
        
        def flags(self, index):
            if not index.isValid():
                return Qt.ItemIsEnabled
            
            editable = Qt.ItemIsEditable if self.header[index.column()].editable else Qt.ItemFlags()
            
            # super() doesn't work here :/
            #return Qt.ItemFlags(super(PyTableModel, self).flags(index) | editable)
            return Qt.ItemFlags(QAbstractTableModel.flags(self, index) | editable)
        
        def calcPosition(self, position):
            if position < 0:
                position += self.rowCount() + 1
                if position < 0:
                    position = 0
            elif position > self.rowCount():
                position = self.rowCount()
            
            return position
        
        def insertRows(self, position, dataList, index=QModelIndex()):
            position = self.calcPosition(position)
            
            if dataList is None or len(dataList) == 0:
                return False
            
            self.beginInsertRows(QModelIndex(), position, position + len(dataList) - 1)
            self.dataList[position:position] = dataList
            self.endInsertRows()
            return True
        
        def removeRows(self, position, rows=1, index=QModelIndex()):
            position = self.calcPosition(position)
            if rows <= 0:
                return False
            
            self.beginRemoveRows(QModelIndex(), position, position + rows - 1)
            del self.dataList[position:position + rows]
            self.endRemoveRows()
            
            return True
        
        def __getitem__(self, index):
            if isinstance(index, QModelIndex):
                index = index.row()
            return self.dataList[index]
    
    class PyAttrColumn(Column):
        def __init__(self, name, attr, **kw_args):
            super(PyAttrColumn, self).__init__(name, **kw_args)
            
            self.attr = attr
        
        def toString(self, obj):
            return getattr(obj, self.attr)
    
    class PyAttrLookupColumn(PyAttrColumn):
        def __init__(self, name, attr, value2TextMap, attr2=None, **kw_args):
            super(PyAttrColumn, self).__init__(name, **kw_args)
            
            self.attr = attr
            self.value2TextMap = value2TextMap
            self.attr2 = attr2
        
        def toString(self, obj):
            try:
                id_ = getattr(obj, self.attr)
                v = self.value2TextMap.get(id_)
                if self.attr2 and v is not None:
                    v = getattr(v, self.attr2)
                return v
            except:
                log.exception('Unable to get attribute %s from %s' % (self.attr, repr(obj)))
                return u''
    
    class SeriesIssueCacheColumn(Column):
        def __init__(self, name, cache, **kw_args):
            super(SeriesIssueCacheColumn, self).__init__(name, **kw_args)
            
            self.cache = cache
        
        def toString(self, obj):
            if obj:
                v = self.cache.get(obj.issue)
                if v is None:
                    return u''
                else:
                    return v.name
            else:
                return u''
        
    class LocationTableModel(PyTableModel):
        def __init__(self, parent=None):
            super(LocationTableModel, self).__init__(parent)
            
            self.addColumn(PyAttrColumn(tr('ID'), 'id', dataAlignment=(Qt.AlignRight|Qt.AlignVCenter)))
            self.addColumn(PyAttrColumn(tr('Name'), 'name'))
            self.addColumn(PyAttrColumn(tr('Barcode'), 'barcode'))
    
    class PySortFilterProxyModel(QSortFilterProxyModel):
        def __init__(self, parent=None):
            super(PySortFilterProxyModel, self).__init__(parent)
        
        def headerData(self, section, orientation, role=Qt.DisplayRole):
            if orientation == Qt.Vertical and role == Qt.DisplayRole:
                return QVariant(int(section + 1))
            
            # super() doesn't work here :/
            #return super(PySortFilterProxyModel, self).headerData(section, orientation, role)
            return QSortFilterProxyModel.headerData(self, section, orientation, role)
        
        def __getitem__(self, index):
            if type(index) == types.IntType:
                index = QSortFilterProxyModel.index(self, index, 0)
            index = self.mapToSource(index)
            return self.sourceModel()[index]
    
    class LocationTable(QTableView):
        def __init__(self, parent=None):
            super(LocationTable, self).__init__(parent)
            
            self.proxyModel = PySortFilterProxyModel()
            self.realModel = LocationTableModel()
            
            self.proxyModel.setSourceModel(self.realModel)
            self.setModel(self.proxyModel)
            
            # TODO setRowHeight
        
        def __getitem__(self, index):
            return self.proxyModel[index]
    
    
    class IssueCache(object):
        def __init__(self, db):
            assert db is not None
            
            self.db = db
            self.cache = {}
        
        def get (self, value):
            issue = self.cache.get(value, None)
            if issue is None and value not in self.cache:
                issue = loadIssue(self.db, value)
                self.cache[value] = issue
            
            return issue
    
    class SeriesCache(object):
        def __init__(self, db, issueCache):
            assert db is not None
            
            self.db = db
            self.issueCache = issueCache
            self.cache = {}
        
        def get (self, value):
            issue = self.issueCache.get(value)
            if issue is None:
                return None
            
            series = self.cache.get(issue.series, None)
            if series is None and issue.series not in self.cache:
                series = loadSeries(self.db, issue.series)
                self.cache[series.id] = series
            return series
    
    class ArticleTableModel(PyTableModel):
        AC_ID, AC_TYPE, AC_TITLE, AC_CREATOR, AC_COMMENT, AC_LOCATION, AC_LENT_TO, AC_SERIES, AC_ISSUE, AC_HAS_IMAGE, AC_HAS_URL = range(11)
        
        def __init__(self, seriesCache, issueCache, parent=None):
            super(ArticleTableModel, self).__init__(parent)
            
            self.addColumn(PyAttrColumn(tr('ID'), 'id', dataAlignment=(Qt.AlignRight|Qt.AlignVCenter)))
            self.addColumn(PyAttrColumn(tr('Type'), 'type'))
            self.addColumn(PyAttrColumn(tr('Title'), 'title'))
            self.addColumn(PyAttrColumn(tr('Author / Artist'), 'creator'))
            self.addColumn(PyAttrColumn(tr('Comment'), 'comment'))
            self.locationColumn = PyAttrLookupColumn(tr('Location'), 'location', {}, 'name')
            self.addColumn(self.locationColumn)
            self.lentToColumn = PyAttrLookupColumn(tr('Lent To'), 'lentTo', {}, 'name')
            self.addColumn(self.lentToColumn)
            self.addColumn(SeriesIssueCacheColumn(tr('Series'), seriesCache))
            self.addColumn(SeriesIssueCacheColumn(tr('Issue'), issueCache))
            yesNoMap = {
                'Yes': tr('Yes'),
                'No': tr('No'),
            }
            self.addColumn(PyAttrLookupColumn(tr('Has Image?'), 'hasImage', yesNoMap))
            self.addColumn(PyAttrLookupColumn(tr('Has URL?'), 'hasURL', yesNoMap))
    
    class ArticleTable(QTableView):
        def __init__(self, seriesCache, issueCache, parent=None):
            super(ArticleTable, self).__init__(parent)
            
            self.proxyModel = PySortFilterProxyModel()
            self.realModel = ArticleTableModel(seriesCache, issueCache)
            
            self.proxyModel.setSourceModel(self.realModel)
            self.setModel(self.proxyModel)
            
            self.setHorizontalScrollMode(QTableView.ScrollPerPixel)
            # TODO setRowHeight
        
        def __getitem__(self, index):
            return self.proxyModel[index]
    
    
    class PyComboBox(QComboBox):
        def __init__(self, parent=None):
            super(PyComboBox, self).__init__(parent)
        
        def select(self, data):
            for i in range(0, self.count()):
                itemData = self.itemData(i).toInt()[0]
                if data == itemData:
                    self.setCurrentIndex(i)
                    return
            self.setCurrentIndex(0)
    
    class MainWindow(QMainWindow):
        def __init__(self, db, parent=None):
            super(MainWindow, self).__init__(parent)
            self.resize(1200, 800)
            self.setWindowTitle(tr(u'UPCScan %1', VERSION))
            
            self.log = logging.getLogger('MainWindow')
            
            global progress
            progress = self.progress
            
            self.coverData = None
            self.coverImage = None
            self.coverImageFormat = ''
            self.coverPixmap = ''
            self.currentLocation = 0
            self.coverDialog = None
            self.lastUpdateRequest = None
            self.lastCoverLoadRequest = None
            
            self.db = db
            self.issueCache = IssueCache(self.db)
            self.seriesCache = SeriesCache(self.db, self.issueCache)
            self.createWidgets()
            
            self.connect(self, SIGNAL('locationsLoaded'), self.addLocations)
            self.loadLocations()
            self.connect(self, SIGNAL('articlesLoaded'), self.addArticles)
            self.loadArticles()
            self.connect(self, SIGNAL('coverImageLoaded'), self.getCoverImageDataFromDBResult)
            
            self.progress(tr('Ready.'))
            self.scanCode.setFocus()

        def progress(self, msg):
            self.statusBar.showMessage(msg)

        def selectArticle(self, row, col):
            index = self.articleModel.index(row, 0, QModelIndex())
            index2 = self.articleModel.index(row, self.articleModel.columnCount()-1, QModelIndex())
            selection = QItemSelection()
            selection.select(index, index2)
            self.articleTable.selectionModel().select(selection, QItemSelectionModel.ClearAndSelect)
            
            index = self.articleModel.index(row, col, QModelIndex())
            self.articleTable.scrollTo(index)

        def countArticles(self):
            query = QSqlQuery("select Count(*) from article", self.db)
            count = 0
            if query.next():
                count, ok = query.value(0).toInt()
            return count

        def nextId(self):
            query = QSqlQuery("select Max(ID) from article", self.db)
            id_ = 1
            if query.next():
                id_, ok = query.value(0).toInt()
                id_ += 1
            return id_

        def createWidgets(self):
            self.central_widget = QWidget()
            vbox = QVBoxLayout()
            self.central_widget.setLayout(vbox)
            self.central_widget.setContentsMargins(0,0,0,0) 
            self.setCentralWidget(self.central_widget)
            vbox.setContentsMargins(0,0,0,0) 
            
            self.locationTable = LocationTable(self)
            self.locationModel = self.locationTable.proxyModel
            vbox.addWidget(self.locationTable)
            
            self.loadLocations()
            self.id2Location = {}
            
            self.locationTable.setSelectionBehavior(QAbstractItemView.SelectRows)
            self.locationTable.setSelectionMode(QAbstractItemView.SingleSelection)
            self.locationTable.setSortingEnabled(True)
            self.locationTable.sortByColumn(0, Qt.AscendingOrder)
            self.locationTable.setColumnWidth(0, 50)
            self.locationTable.setColumnWidth(1, 250)
            self.locationTable.setColumnWidth(2, 100)
            
            selectionModel = self.locationTable.selectionModel()
            self.connect(selectionModel, SIGNAL('selectionChanged(const QItemSelection &,const QItemSelection &)'), self.locationSelectionChanged)
            
            self.articleTable = ArticleTable(self.seriesCache, self.issueCache, self)
            vbox.addWidget(self.articleTable)
            self.articleModel = self.articleTable.proxyModel
            self.articleTable.setSelectionBehavior(QAbstractItemView.SelectRows)
            self.articleTable.setSelectionMode(QAbstractItemView.SingleSelection)
            self.articleTable.realModel.locationColumn.value2TextMap = self.id2Location
            self.articleTable.realModel.lentToColumn.value2TextMap = self.id2Location
            self.articleTable.setSortingEnabled(True)
            self.articleTable.sortByColumn(0, Qt.AscendingOrder)
            self.articleTable.setColumnWidth(0, 50)
            self.articleTable.setColumnWidth(1, 80)
            self.articleTable.setColumnWidth(2, 250)
            self.articleTable.setColumnWidth(3, 250)
            self.articleTable.setColumnWidth(4, 250)
            
            selectionModel = self.articleTable.selectionModel()
            self.connect(selectionModel, SIGNAL('selectionChanged(const QItemSelection &,const QItemSelection &)'), self.articleSelectionChanged)
            
            grid = QGridLayout()
            grid.setHorizontalSpacing(4)
            grid.setVerticalSpacing(0)
            grid.setColumnStretch(2, 1)
            grid.setColumnStretch(4, 1)
            vbox.addLayout(grid)
            
            row = 0
            label = QLabel(tr('Scan Location:'))
            grid.addWidget(label, row, 1)
            self.currentLocationWidget = QLabel()
            grid.addWidget(self.currentLocationWidget, row, 2, 1, 3)

            editButton = QPushButton()
            editButton.setText(tr('&Edit...'))
            editButton.setToolTip(tr('Edit the available locations'))
            self.connect(editButton, SIGNAL('clicked(bool)'), self.editLocations)
            grid.addWidget(editButton, row, 5)
            row += 1
            
            label = QLabel(tr('&Scan Code:'))
            grid.addWidget(label, row, 1)
            self.scanCode = QLineEdit()
            grid.addWidget(self.scanCode, row, 2, 1, 3)
            self.connect(self.scanCode, SIGNAL('editingFinished()'), self.scan)
            label.setBuddy(self.scanCode)
            row += 1
            
            # Empty row here
            grid.setRowMinimumHeight(row, 15)
            row += 1
            
            hbox = QHBoxLayout()
            grid.addLayout(hbox, row, 0, 6, 1)
            hbox.setContentsMargins(5, 0, 5, 0)
            self.coverImageWidget = QPushButton()
            self.coverImageWidget.setText(tr('Cover Image'))
            self.coverImageWidget.setToolTip(tr('The cover image of this product'))
            self.coverImageWidget.setIconSize(QSize(128, 128))
            self.coverImageWidget.setMinimumSize(QSize(128, 128))
            self.coverImageWidget.resize(QSize(128, 128))
            self.connect(self.coverImageWidget, SIGNAL('clicked(bool)'), self.coverClicked)
            hbox.addWidget(self.coverImageWidget)
            
            label = QLabel(tr('T&ype:'))
            grid.addWidget(label, row, 1)
            self.typeWidget = QLineEdit()
            self.typeWidget.setToolTip(tr('What is this? A book? DVD? CD? Comic? Magazine?'))
            grid.addWidget(self.typeWidget, row, 2, 1, 3)
            self.connect(self.typeWidget, SIGNAL('textEdited(const QString &)'), self.typeChanged)
            label.setBuddy(self.typeWidget)
            
            self.addButton = QPushButton()
            self.addButton.setText(tr('Add'))
            self.addButton.setToolTip(tr('Add a new article to the list'))
            self.connect(self.addButton, SIGNAL('clicked(bool)'), self.addClicked)
            grid.addWidget(self.addButton, row, 5)
            row += 1
            
            label = QLabel(tr('&Title:'))
            grid.addWidget(label, row, 1)
            self.titleWidget = QLineEdit()
            self.titleWidget.setToolTip(tr('The name or title of the product'))
            self.connect(self.titleWidget, SIGNAL('textEdited(const QString &)'), self.titleChanged)
            grid.addWidget(self.titleWidget, row, 2, 1, 3)
            label.setBuddy(self.titleWidget)
            
            self.deleteButton = QPushButton()
            self.deleteButton.setText(tr('Delete...'))
            self.deleteButton.setToolTip(tr('Delete the selected article(s)'))
            self.connect(self.deleteButton, SIGNAL('clicked(bool)'), self.deleteClicked)
            grid.addWidget(self.deleteButton, row, 5)
            row += 1
            
            label = QLabel(tr('&Creator:'))
            grid.addWidget(label, row, 1)
            self.creatorWidget = QLineEdit()
            self.creatorWidget.setToolTip(tr('Name of the author, artist, actor of this book, music CD or movie'))
            grid.addWidget(self.creatorWidget, row, 2, 1, 3)
            self.connect(self.creatorWidget, SIGNAL('textEdited(const QString &)'), self.creatorChanged)
            label.setBuddy(self.creatorWidget)
            row += 1
            
            label = QLabel(tr('C&omment:'))
            grid.addWidget(label, row, 1)
            self.commentWidget = QLineEdit()
            self.commentWidget.setToolTip(tr('Your personal comment'))
            grid.addWidget(self.commentWidget, row, 2, 1, 3)
            self.connect(self.commentWidget, SIGNAL('textEdited(const QString &)'), self.commentChanged)
            label.setBuddy(self.commentWidget)
            row += 1
            
            label = QLabel(tr('&URL:'))
            grid.addWidget(label, row, 1)
            self.urlWidget = QLineEdit()
            self.urlWidget.setToolTip(tr('The URL of a website which has more information about this product'))
            grid.addWidget(self.urlWidget, row, 2, 1, 3)
            self.connect(self.urlWidget, SIGNAL('textEdited(const QString &)'), self.urlChanged)
            label.setBuddy(self.urlWidget)
            self.visitButton = QPushButton()
            self.visitButton.setText(tr('&Visit'))
            self.visitButton.setToolTip(tr('Visit the web site at the URL to the left'))
            self.connect(self.visitButton, SIGNAL('clicked(bool)'), self.visitClicked)
            grid.addWidget(self.visitButton, row, 5)
            row += 1
            
            label = QLabel(tr('&Series:'))
            grid.addWidget(label, row, 1)
            self.seriesWidget = QComboBox()
            self.seriesWidget.setToolTip(tr('The series (TV, comic book, etc) this article belongs to'))
            grid.addWidget(self.seriesWidget, row, 2)
            label.setBuddy(self.seriesWidget)
            self.connect(self.seriesWidget, SIGNAL('activated(int)'), self.seriesChanged)
            
            label = QLabel(tr('&Issue:'))
            grid.addWidget(label, row, 3)
            self.issueWidget = QComboBox()
            self.issueWidget.setToolTip(tr('The issue in the series'))
            grid.addWidget(self.issueWidget, row, 4)
            label.setBuddy(self.issueWidget)
            self.connect(self.issueWidget, SIGNAL('activated(int)'), self.issueChanged)
            
            editButton = QPushButton()
            editButton.setText(tr('Edit...'))
            editButton.setToolTip(tr('Edit the available series/issues'))
            self.connect(editButton, SIGNAL('clicked(bool)'), self.editSeries)
            grid.addWidget(editButton, row, 5)
            row += 1
            
            label = QLabel(tr('&Location:'))
            grid.addWidget(label, row, 1)
            self.locationWidget = PyComboBox()
            self.locationWidget.setToolTip(tr('Where you can usually find this product. Is it in a room, a shelve or somewhere else?'))
            grid.addWidget(self.locationWidget, row, 2)
            label.setBuddy(self.locationWidget)
            self.connect(self.locationWidget, SIGNAL('activated(int)'), self.locationChanged)

            label = QLabel(tr('&Lent to:'))
            grid.addWidget(label, row, 3)
            self.lentToWidget = PyComboBox()
            self.lentToWidget.setToolTip(tr('Where you can currently find this product. Change this if you borrowed it to a soon to-be ex-friend?'))
            grid.addWidget(self.lentToWidget, row, 4)
            label.setBuddy(self.lentToWidget)
            self.connect(self.lentToWidget, SIGNAL('activated(int)'), self.lentToChanged)
            
            editButton = QPushButton()
            editButton.setText(tr('Edit...'))
            editButton.setToolTip(tr('Edit the available locations'))
            self.connect(editButton, SIGNAL('clicked(bool)'), self.editLocations)
            grid.addWidget(editButton, row, 5)
            row += 1
            
            self.statusBar = QStatusBar()
            vbox.addWidget(self.statusBar)
        
        def loadLocations(self):
            #self.log.debug('loadLocations')
            loader = ObjectLoader(Location, Location.ALL_COLUMNS)
            loader.callback = self.locationsLoaded
            # Start loading of objects in a background thread
            self.db.execute(loader)
        
        def locationsLoaded(self, l):
            #self.log.debug('Emitting locationsLoaded')
            # This gets called from a background thread. Send a signal to the
            # UI thread that the objects are now ready.
            self.emit(SIGNAL('locationsLoaded'), l)
        
        def addLocations(self, l):
            # Add the locations to the view. This must be run from the UI thread.
            #self.log.debug('Adding locations to UI')
            self.locationTable.realModel.insertRows(-1, l)
            for loc in l:
                self.id2Location[loc.id] = loc
            
            l = self.id2Location.values()
            l.sort(key=lambda item: item.name.lower())
            
            self.locationWidget.addItem(u'-', QVariant(-1))
            self.lentToWidget.addItem(u'-', QVariant(-1))
            for loc in l:
                self.locationWidget.addItem(loc.name, QVariant(loc.id))
                self.lentToWidget.addItem(loc.name, QVariant(loc.id))
            
            if self.currentLocation is None and self.locationModel.rowCount() > 0:
                index = self.locationModel.index(0, 0, QModelIndex())
                index2 = self.locationModel.index(0, self.locationModel.columnCount()-1, QModelIndex())
                selection = QItemSelection()
                selection.select(index, index2)
                self.locationTable.selectionModel().select(selection, QItemSelectionModel.ClearAndSelect)
        
        def loadArticles(self, **where_cond):
            #self.log.debug('loadArticles')
            
            loader = ObjectLoader(Article,
                ('id', 'type', 'title', 'creator', 'location', 'comment', 'detailurl', 'asin', 'hasImage', 'hasURL', 'lentTo', 'issue'),
                **where_cond
            )
            
            sql = loader.sql
            sql = sql.replace('hasImage', "case when image notnull then 'Yes' ELSE 'No' end", 1)
            sql = sql.replace('hasURL', "case when detailurl notnull then 'Yes' ELSE 'No' end", 1)
            loader.sql = sql
            
            loader.callback = self.articlesLoaded
            self.db.execute(loader)
        
        def articlesLoaded(self, l):
            #self.log.debug('Emitting articlesLoaded')
            self.emit(SIGNAL('articlesLoaded'), l)
        
        def addArticles(self, l):
            self.articleTable.realModel.insertRows(-1, l)
            
            #print self.articleModel.rowCount()
            if self.articleModel.rowCount() > 0:
                self.selectArticle(self.articleModel.rowCount()-1, 0)
            else:
                self.articleSelectionChanged(QItemSelection(), QItemSelection())
        
        def switchLocation(self, key):
            #print 'switchLocation',key
            for i in range(self.locationTable.model().rowCount()):
                loc = self.locationTable[i]
                if loc.barcode == key:
                    self.setLocation(loc)
                    self.progress(tr(u'Switching to location %1', loc.name))
                    return
            
            self.progress(tr(u'Unknown location %1', key))
        
        def scan(self):
            key = unicode(self.scanCode.text())
            self.scanCode.setText('')
            if not key.strip():
                return
            
            if key[:4] == 'F200': # For scanners which return the barcode type
                key = key[1:]
            if key[:3] == '200':
                self.switchLocation(key)
                return

            req = DBRequest("select id from article where searchKey = ?", (key,))
            rows = self.db.executeAndWait(req)
            if len(rows) > 0:
                self.articleAlreadyExists(rows[0][0], key)
                return

            try:
                root, response = fetcher.fetch(key)
                a = createArticle(root, response)
            except Exception, e:
                self.log.exception('Error while searching article %s' % repr(key))
                progress(tr('Article %1 not found anywhere.', key))
                #with open('failed-codes', 'a') as fh:
                #    fh.write('%s %s\n' % (key, repr(e)))
                a = Article()
                a.asin = 'Unknown-%s' % key
                a.title = 'Unknown %s' % key
                a.type = 'Unknown'
                a.searchType = 'Unknown'
                a.searchKey = key
                                        
            req = DBRequest("select id from article where asin = ?", (a.asin,))
            rows = self.db.executeAndWait(req)
            if len(rows) > 0:
                self.articleAlreadyExists(rows[0][0], key)
                return
            
            a.location = self.currentLocation.id if self.currentLocation else None
            #print a
            
            issue = findIssue(self.db, a.asin)
            if issue is not None:
                a.issue = issue.id
            
            self.insert(a)
            
            self.progress(tr('Ready.'))
        
        def articleAlreadyExists(self, id_, key):
            for i in range(self.articleTable.model().rowCount()):
                a = self.articleTable[i]
                if a.id == id_:
                    self.selectArticle(i, 0)
                    break
            self.statusBar.showMessage(tr('Article "%1" already exists', key))
        
        def insert(self, a):
            insertObject(self.db, a, Article.ALL_COLUMNS)
            self.loadArticles(asin=a.asin)
            #l = loadArticles(self.cursor, asin=a.asin)
            #self.addArticles(l)

        def typeChanged(self, text):
            self.updateCurrentArticle(ArticleTableModel.AC_TYPE, 'type', text, self.typeWidget)
        
        def titleChanged(self, text):
            self.updateCurrentArticle(ArticleTableModel.AC_TITLE, 'title', text, self.titleWidget)
        
        def creatorChanged(self, text):
            self.updateCurrentArticle(ArticleTableModel.AC_CREATOR, 'creator', text, self.creatorWidget)
        
        def urlChanged(self, text):
            self.updateCurrentArticle(-1, 'detailurl', text, self.urlWidget)
        
        def commentChanged(self, text):
            self.updateCurrentArticle(ArticleTableModel.AC_COMMENT, 'comment', text, self.commentWidget)
        
        def seriesChanged(self, index):
            # TODO fill issue combo box
            self.progress(tr(u'Not implemented'))
        
        def issueChanged(self, index):
            # TODO
            self.progress(tr(u'Not implemented'))
        
        def locationChanged(self, index):
            loc, ok = self.locationWidget.itemData(index).toInt()
            assert ok
            self.updateCurrentArticle(ArticleTableModel.AC_ISSUE, 'location', loc, self.locationWidget)
        
        def lentToChanged(self, index):
            loc, ok = self.lentToWidget.itemData(index).toInt()
            assert ok
            self.updateCurrentArticle(ArticleTableModel.AC_LENT_TO, 'lentTo', loc, self.lentToWidget)
        
        def updateCurrentArticle(self, col, attr, value, widget):
            if self.currentArticle is None:
                raise Exception('self.currentArticle is None')
            
            if attr in ('location', 'lentTo'):
                if value < 0:
                    value = None
            else:
                # IMPORTANT If you forget this line, you'll get ugly crashes
                value = unicode(value)
            
            setattr(self.currentArticle, attr, value)
            
            if col != -1:
                model = self.articleTable.model()
                index = model.index(self.currentArticleIndex.row(), col, QModelIndex())
                model.setData(index, QVariant(value))
            
            attrList = (attr,)
            if self.lastUpdateRequest is not None \
                and self.lastUpdateRequest.obj == self.currentArticle \
                and self.lastUpdateRequest.fields == attrList:
                self.lastUpdateRequest.cancel.set()
            
            self.lastUpdateRequest = DBUpdateRequest(self.currentArticle, attrList)
            self.db.execute(self.lastUpdateRequest)

        def locationSelectionChanged(self, selected, deselected):
            if len(selected.indexes()) > 0:
                index = selected.indexes()[0]
                self.setLocation(self.locationModel[index])
        
        def setLocation(self, loc):
            self.currentLocation = loc
            self.currentLocationWidget.setText(loc.name)
        
        def articleSelectionChanged(self, selected, deselected):
            if len(selected.indexes()) > 0:
                index = selected.indexes()[0]
                a = self.articleTable[index]
                self.currentArticle = a
                self.currentArticleIndex = index
                
                def model2Edit(attr, widget):
                    v = getattr(a, attr)
                    widget.setText('' if v is None else v)
                    widget.setEnabled(True)
                
                model2Edit('type', self.typeWidget)
                model2Edit('title', self.titleWidget)
                model2Edit('creator', self.creatorWidget)
                model2Edit('detailurl', self.urlWidget)
                model2Edit('comment', self.commentWidget)
                
                self.locationWidget.select(a.location)
                self.lentToWidget.select(a.lentTo)
                
                # Make sure the start of the URL is visible
                self.urlWidget.setCursorPosition(0)
                # TODO Set seriesWidget
                # TODO Set issueWidget
            else:
                def disable(widget):
                    widget.setText('')
                    widget.setEnabled(False)
                
                disable(self.typeWidget)
                disable(self.titleWidget)
                disable(self.creatorWidget)
                disable(self.commentWidget)
                disable(self.urlWidget)
                self.currentArticle = None
                self.currentArticleIndex = -1
                self.seriesWidget.setCurrentIndex(0)
                self.issueWidget.setCurrentIndex(0)
                self.locationWidget.setCurrentIndex(0)
                self.lentToWidget.setCurrentIndex(0)
                self.coverData = None
                self.coverImageWidget.setText('')
                self.coverImageWidget.setIcon(QIcon())
                self.coverImageWidget.setEnabled(False)
            
            enabled = (self.currentArticle is not None)
            self.visitButton.setEnabled(enabled)
            self.deleteButton.setEnabled(enabled)
            self.seriesWidget.setEnabled(enabled)
            self.issueWidget.setEnabled(enabled)
            self.locationWidget.setEnabled(enabled)
            self.lentToWidget.setEnabled(enabled)
            
            self.loadCoverWidget()
        
        def loadCoverWidget(self):
            if self.lastCoverLoadRequest is not None:
                self.lastCoverLoadRequest.cancel.set()
            
            if self.currentArticle is not None:
                loader = DBRequest("select image from article where id = ?", (
                    self.currentArticle.id,
                ))
                loader.callback = self.coverImageLoaded
                self.lastCoverLoadRequest = loader
                self.db.execute(loader)
            
        def coverImageLoaded(self, l):
            self.emit(SIGNAL('coverImageLoaded'), l)
        
        def getCoverImageDataFromDBResult(self, rows):
            self.lastCoverLoadRequest = None
            self.coverData = None
            if len(rows) == 1 and rows[0][0] is not None:
                self.coverData = rows[0][0]
            self.updateCoverWidget()
        
        def updateCoverWidget(self):
            self.coverImage = None
            self.coverPixmap = None
            self.coverImageFormat = ''
            
            if self.coverData is not None:
                self.coverImage, self.coverImageFormat = bufferToImage(self.coverData)
            
            if self.coverImage is not None:
                self.coverPixmap = QPixmap.fromImage(self.coverImage)
            
            
            text = u''
            hasImage = 'Yes'
            if self.coverPixmap is None:
                text = tr('No Cover')
                icon = QIcon()
                hasImage = 'No'
            else:
                icon = QIcon(self.coverPixmap)
                
            self.coverImageWidget.setText(text)
            self.coverImageWidget.setIcon(icon)
            self.coverImageWidget.setEnabled(True)
            
            if self.currentArticle.hasImage != hasImage:
                self.currentArticle.hasImage = hasImage
                model = self.articleTable.model()
                index = model.index(self.currentArticleIndex.row(), ArticleTableModel.AC_HAS_IMAGE, QModelIndex())
                model.setData(index, QVariant(hasImage))
            
            if self.coverDialog is not None:
                self.coverDialog.updateCoverWidget()
        
        def coverClicked(self, checked):
            if self.coverDialog is None:
                self.coverDialog = CoverDialog(self)
            
            self.coverDialog.show()
            
            if self.coverImage:
                self.coverDialog.updateCoverWidget()
        
        def visitClicked(self, checked):
            self.progress(tr('Starting browser...'))
            import webbrowser
            webbrowser.open(self.urlWidget.text())

        def editLocations(self, checked):
            # TODO open a dialog to edit the locations
            self.progress(tr('Not implemented.'))
        
        def editSeries(self, checked):
            # TODO Open dialog to edit the series/issues
            self.progress(tr('Not implemented.'))
        
        def deleteClicked(self, checked):
            if self.currentArticle is None:
                return
            
            index = self.currentArticleIndex
            a = self.currentArticle
            ret = QMessageBox.warning(self, 
                tr(u'Are you sure?'),
                tr(u'Are you sure to delete\n"%1"?', a.title),
                QMessageBox.Yes | QMessageBox.No,
                QMessageBox.No
            )
            if ret != QMessageBox.Yes:
                return
            
            self.currentArticle = None
            self.currentArticleIndex = -1
            
            a.delete(self.db)
            index = index.row()
            self.articleTable.proxyModel.removeRows(index, 1)
            
            if index > 0:
                self.selectArticle(index-1, 0)
        
        def addClicked(self, checked):
            a = Article()
            import uuid
            a.asin = 'Private-%s' % uuid.uuid4()
            a.type = 'Please specify the type'
            a.title = 'Title of your article'
            a.searchType = 'Manual'
            a.searchKey = 'Manual'
            
            a.insert(self.db)
            self.addArticles([a])

    app = QApplication(sys.argv)
    mainWindow = MainWindow(db)
    mainWindow.show()
    
    app.exec_()

def export(filename):
    with codecs.open(filename, 'w', 'utf-8') as fh:
        fh.write('<?xml version="1.0" encoding="utf-8"?>\n\n')
        fh.write('<upcscan>\n')
        
        db = DB(dbName)
        try:
            db.start()
            
            fh.write('\t<locations>\n')
            
            req = DBExportRequest('select %s from %s order by id' % (
                    ', '.join(Location.ALL_COLUMNS), Location.TABLE_NAME,
                ),
                None,
                fh,
                'location',
                Location.ALL_COLUMNS
            )
            db.execute(req)
            req.done.wait()
            
            fh.write('\t</locations>\n')
            
            fh.write('\t<articles>\n')
            
            req = DBExportRequest('select %s from %s order by id' % (
                    ', '.join(Article.ALL_COLUMNS), Article.TABLE_NAME,
                ),
                None,
                fh,
                'article',
                Article.ALL_COLUMNS
            )
            db.execute(req)
            req.done.wait()
            
            fh.write('\t</articles>\n')
            
        finally:
            db.close()
    
        fh.write('</upcscan>\n')
    
if __name__ == '__main__':
    log = logging.getLogger('main')
    try:
        if len(sys.argv) > 1 and sys.argv[1] == '-t':
            runTests()
        elif len(sys.argv) > 1 and sys.argv[1] == '-e':
            if len(sys.argv) < 3:
                raise Exception('Missing export filename')
            filename = sys.argv[2]
            export(filename)
        else:
            upgradeDB(dbName)
            
            needToCreateDB = not os.path.exists(dbName)
            
            db = DB(dbName)
            db.start()
            
            if needToCreateDB:
                createDB(db)
            
            try:
                
                hasQt = False
                if len(sys.argv) > 1 and sys.argv[1] == '-s':
                    log.debug('Shell mode enforced')
                else:
                    log.debug('Checking for PyQt4')
                    try:
                        from PyQt4.pyqtconfig import Configuration
                        conf = Configuration()
                        hasQt = True
                        qt_version = '%d.%d.%d' % (
                            ((conf.qt_version & 0xff0000) >> 16),
                            ((conf.qt_version & 0x00ff00) >> 8),
                            ((conf.qt_version & 0x0000ff) >> 0),
                        )
                        log.info('PyQt4 %s (Qt %s) found.' % (conf.pyqt_version_str, qt_version))
                    except:
                        log.info('No PyQt4 found. Running in console mode.')
                    
                if hasQt:
                    runQt4(db)
                else:
                    run(db)
            finally:
                db.close()
    except:
        log.exception('Execution failed.')
    
